#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Parses the Mbed configuration system and generates a CMake config script."""
import pathlib

from mbed_tools.project import MbedProgram
from mbed_tools.targets import get_target_by_name
from mbed_tools.build._internal.cmake_file import generate_mbed_config_cmake_file
from mbed_tools.build._internal.config.assemble_build_config import assemble_config
from mbed_tools.build._internal.write_files import write_file


def generate_config(target_name: str, toolchain: str, program: MbedProgram) -> pathlib.Path:
    """Generate an Mbed config file at the program root by parsing the mbed config system.

    Args:
        target_name: Name of the target to configure for.
        toolchain: Name of the toolchain to use.
        program: The MbedProgram to configure.

    Returns:
        Path to the generated config file.
    """
    target_build_attributes = get_target_by_name(target_name, program.mbed_os.targets_json_file)
    config = assemble_config(target_build_attributes, program.root, program.files.app_config_file)
    cmake_file_contents = generate_mbed_config_cmake_file(target_name, target_build_attributes, config, toolchain)
    cmake_config_file_path = program.files.cmake_config_file
    write_file(cmake_config_file_path.parent, cmake_config_file_path.name, cmake_file_contents)
    return cmake_config_file_path



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import os
from unittest import TestCase, mock

from mbed_tools.targets.env import env


class TestMbedApiAuthToken(TestCase):
    @mock.patch.dict(os.environ, {"MBED_API_AUTH_TOKEN": "sometoken"})
    def test_returns_api_token_set_in_env(self):
        self.assertEqual(env.MBED_API_AUTH_TOKEN, "sometoken")


class TestDatabaseMode(TestCase):
    @mock.patch.dict(os.environ, {"MBED_DATABASE_MODE": "ONLINE"})
    def test_returns_database_mode_set_in_env(self):
        self.assertEqual(env.MBED_DATABASE_MODE, "ONLINE")

    def test_returns_default_database_mode_if_not_set_in_env(self):
        self.assertEqual(env.MBED_DATABASE_MODE, "AUTO")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Loads system data in parallel and all at once in order to improve performance."""
from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Dict, Generator, Optional, cast

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptorWrapper, ComponentDescriptor
from mbed_tools.devices._internal.windows.disk_drive import DiskDrive
from mbed_tools.devices._internal.windows.disk_partition import DiskPartition
from mbed_tools.devices._internal.windows.disk_partition_logical_disk_relationships import (
    DiskPartitionLogicalDiskRelationship,
)
from mbed_tools.devices._internal.windows.logical_disk import LogicalDisk
from mbed_tools.devices._internal.windows.serial_port import SerialPort
from mbed_tools.devices._internal.windows.usb_controller import UsbController
from mbed_tools.devices._internal.windows.usb_hub import UsbHub

# All Windows system data of interest in order to retrieve the information for DeviceCandidate.
SYSTEM_DATA_TYPES = [
    UsbHub,
    UsbController,
    DiskDrive,
    DiskPartition,
    LogicalDisk,
    DiskPartitionLogicalDiskRelationship,
    SerialPort,
]


def load_all(cls: type) -> Tuple[type, List[ComponentDescriptor]]:
    """Loads all elements present in the system referring to a specific type."""
    return (cls, [element for element in ComponentDescriptorWrapper(cls).element_generator()])


class SystemDataLoader:
    """Object in charge of loading all system data with regards to Usb, Disk or serial port.

    It loads all the data in parallel and all at once in order to improve performance.
    """

    def __init__(self) -> None:
        """Initialiser."""
        self._system_data: Optional[Dict[type, List[ComponentDescriptor]]] = None

    def _load(self) -> None:
        """Loads all system data in parallel."""
        with ThreadPoolExecutor() as executor:
            results = executor.map(load_all, SYSTEM_DATA_TYPES)
        self._system_data = {k: v for (k, v) in results}

    @property
    def system_data(self) -> Dict[type, List[ComponentDescriptor]]:
        """Gets all system data."""
        if not self._system_data:
            self._load()
        return cast(Dict[type, List[ComponentDescriptor]], self._system_data)

    def get_system_data(self, cls: type) -> List[ComponentDescriptor]:
        """Gets the system data for a particular type."""
        return self.system_data.get(cls, list())


class ComponentsLoader:
    """Loads system components."""

    def __init__(self, data_loader: SystemDataLoader, cls: type) -> None:
        """initialiser."""
        self._cls = cls
        self._data_loader = data_loader

    def element_generator(self) -> Generator["ComponentDescriptor", None, None]:
        """Gets a generator over all elements currently registered in the system."""
        for component in self._data_loader.get_system_data(self._cls):
            yield component



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Resolve targets for `CandidateDevice`.

Resolving a target involves looking up an `MbedTarget` from the `mbed-targets` API, using data found in the "htm file"
located on an "Mbed Enabled" device's USB MSD.

For more information on the mbed-targets package visit https://github.com/ARMmbed/mbed-targets
"""
import itertools
import logging
import pathlib

from typing import Iterable, List, Optional

from mbed_tools.targets import Board, get_board_by_product_code, get_board_by_online_id
from mbed_tools.targets.exceptions import UnknownBoard

from mbed_tools.devices._internal.htm_file import OnlineId, read_online_id, read_product_code
from mbed_tools.devices._internal.candidate_device import CandidateDevice
from mbed_tools.devices._internal.exceptions import NoBoardForCandidate


logger = logging.getLogger(__name__)


def resolve_board(candidate: CandidateDevice) -> Board:
    """Resolves board for a given CandidateDevice.

    This function interrogates CandidateDevice, attempting to establish the best method to resolve a Board,
    the rules are as follows:

    1. Use product code retrieved from one of HTM files in the mass storage if available.
    2. Use online id retrieved from one of the HTM files in the mass storage if available.
    3. Fallback to product code retrieved from serial number.

    The specification of HTM files is that they redirect to board's product page on os.mbed.com.
    Information about Mbed Enabled requirements: https://www.mbed.com/en/about-mbed/mbed-enabled/requirements/
    """
    all_files_contents = _get_all_htm_files_contents(candidate.mount_points)

    product_code = _extract_product_code(all_files_contents)
    if product_code:
        try:
            return get_board_by_product_code(product_code)
        except UnknownBoard:
            logger.error(f"Could not identify a board with the product code: '{product_code}'.")
            raise NoBoardForCandidate

    online_id = _extract_online_id(all_files_contents)
    if online_id:
        slug = online_id.slug
        target_type = online_id.target_type
        try:
            return get_board_by_online_id(slug=slug, target_type=target_type)
        except UnknownBoard:
            logger.error(f"Could not identify a board with the slug: '{slug}' and target type: '{target_type}'.")
            raise NoBoardForCandidate

    # Product code might be the first 4 characters of the serial number
    try:
        product_code = candidate.serial_number[:4]
        return get_board_by_product_code(product_code)
    except UnknownBoard:
        # Most devices have a serial number so this may not be a problem
        logger.info(
            f"The device with the Serial Number: '{candidate.serial_number}' (Product Code: '{product_code}') "
            f"does not appear to be an Mbed development board."
        )
        raise NoBoardForCandidate


def _extract_product_code(all_files_contents: Iterable[str]) -> Optional[str]:
    """Return first product code found in files contents, None if not found."""
    for contents in all_files_contents:
        product_code = read_product_code(contents)
        if product_code:
            return product_code
    return None


def _extract_online_id(all_files_contents: Iterable[str]) -> Optional[OnlineId]:
    """Return first online id found in files contents, None if not found."""
    for contents in all_files_contents:
        online_id = read_online_id(contents)
        if online_id:
            return online_id
    return None


def _get_all_htm_files_contents(directories: Iterable[pathlib.Path]) -> List[str]:
    """Yields all htm files contents found in the list of given directories."""
    files_in_each_directory = (directory.iterdir() for directory in directories)
    all_files = itertools.chain.from_iterable(files_in_each_directory)
    return _read_htm_file_contents(all_files)


def _read_htm_file_contents(all_files: Iterable[pathlib.Path]) -> List[str]:
    htm_files_contents = []
    for file in all_files:
        if _is_htm_file(file):
            try:
                htm_files_contents.append(file.read_text())
            except OSError:
                logger.warning(f"The file '{file}' could not be read from the device, target may not be identified.")
    return htm_files_contents


def _is_htm_file(file: pathlib.Path) -> bool:
    """Checks whether the file looks like an Mbed HTM file."""
    extensions = [".htm", ".HTM"]
    return file.suffix in extensions and not file.name.startswith(".")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Helpers for python language related functions."""
import sys
from collections import namedtuple
from typing import NamedTuple, Iterable, Any, cast


def minimum_python_version(major: int, minor: int) -> bool:
    """States whether the python in use is more recent than set major,minor versions."""
    return sys.version_info.major >= major and sys.version_info.minor >= minor


def named_tuple_with_defaults(typename: str, field_names: Iterable[str], defaults: Iterable[Any]) -> NamedTuple:
    """Gets a named tuple set with default values.

    defaults field only appeared in Python 3.7.
    See https://stackoverflow.com/questions/11351032/namedtuple-and-default-values-for-optional-keyword-arguments
    """
    if minimum_python_version(3, 7):
        return namedtuple(typename=typename, field_names=field_names, defaults=defaults)  # type: ignore
    else:
        a_tuple: NamedTuple = namedtuple(typename=typename, field_names=field_names)  # type: ignore
        a_tuple.__new__.__defaults__ = tuple(defaults)  # type: ignore
        return cast(NamedTuple, a_tuple)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Configure and build a CMake project."""
import logging
import pathlib
import subprocess

from typing import Optional

from mbed_tools.build.exceptions import MbedBuildError


logger = logging.getLogger(__name__)


def build_project(build_dir: pathlib.Path, target: Optional[str] = None) -> None:
    """Build a project using CMake to invoke Ninja.

    Args:
        build_dir: Path to the CMake build tree.
        target: The CMake target to build (e.g 'install')
    """
    target_flag = ["--target", target] if target is not None else []
    _cmake_wrapper("--build", str(build_dir), *target_flag)


def generate_build_system(source_dir: pathlib.Path, build_dir: pathlib.Path, profile: str) -> None:
    """Configure a project using CMake.

    Args:
        source_dir: Path to the CMake source tree.
        build_dir: Path to the CMake build tree.
        profile: The Mbed build profile (develop, debug or release).
    """
    _cmake_wrapper("-S", str(source_dir), "-B", str(build_dir), "-GNinja", f"-DCMAKE_BUILD_TYPE={profile}")


def _cmake_wrapper(*cmake_args: str) -> None:
    try:
        logger.debug("Running CMake with args: %s", cmake_args)
        subprocess.run(["cmake", *cmake_args], check=True)
    except subprocess.CalledProcessError:
        raise MbedBuildError("CMake invocation failed!")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Facilitates parsing MBED.HTM files found on the Mbed devices.

For legacy reasons, we support many flavours of MBED.HTM files. The list of examples below is not exhaustive,
as at the time of writing it's not feasible to gather all examples in one place.
Its main goal is to provide a taste of different flavours of HTM files we need to support.

    <!-- mbed Microcontroller Website and Authentication Shortcut -->
    <!-- Version: 0200 Build: Feb  3 2014 14:03:10 -->
    <html>
    <head>
    <meta http-equiv="refresh" content="0; url=http://mbed.org/device/?code=0710020092566DA9AD33FB84"/>
    <title>mbed Website Shortcut</title>
    </head>
    <body></body>
    </html>

---

    <!doctype html>
    <!-- mbed Platform Website and Authentication Shortcut -->
    <html>
    <head>
    <meta charset="utf-8">
    <title>mbed Website Shortcut</title>
    </head>
    <body>
    <script>
    window.location.replace("https://mbed.org/device/?code=460000000988254a00000000000000000000000097969902?version=0253?target_id=00000000000000000000000000000000");
    </script>
    </body>
    </html>

---

    <!-- mbed Microcontroller Website and Authentication Shortcut -->
    <html>
    <head>
    <meta http-equiv="refresh" content="0; url=http://mbed.org/start?auth=101000000000000000000002F7F35E602eeb0bb9b632205c51f6c357aeee7bc9&loader=11972&firmware=16457&configuration=4" />  # noqa
    <title>mbed Website Shortcut</title>
    </head>
    <body></body>
    </html>

---

    <!doctype html>
    <!-- mbed Platform Website and Authentication Shortcut -->
    <html>
    <head>
    <meta charset="utf-8">
    <title>mbed Website Shortcut</title>
    </head>
    <body>
    <script>
    window.location.replace("https://os.mbed.com/platforms/LPCXpresso54114/");
    </script>
    </body>
    </html>

"""
import re
from typing import Optional, NamedTuple


class OnlineId(NamedTuple):
    """Used to identify the target against the os.mbed.com website.

    The target type and slug are used in the URI for the board and together they can be used uniquely identify a board.

    OnlineId(target_type="platform", slug="SOME-SLUG") -> https://os.mbed.com/platforms/SOME-SLUG
    """

    target_type: str
    slug: str


def read_product_code(file_contents: str) -> Optional[str]:
    """Returns product code parsed from the file contents, None if not found."""
    regex = r"""
            (?:code|auth)=                   # attribute name
            (?P<product_code>[a-fA-F0-9]{4}) # product code
    """
    match = re.search(regex, file_contents, re.VERBOSE)
    if match:
        return match["product_code"]
    return None


def read_online_id(file_contents: str) -> Optional[OnlineId]:
    """Returns online id parsed from the files contents, None if not found."""
    regex = r"""
            (?P<target_type>module|platform)s   # module|platform
            \/                                  # forward slash in the url
            (?P<slug>[-\w]+)                    # permitted characters in a slug are letters and digits
    """
    match = re.search(regex, file_contents, re.VERBOSE)
    if match:
        return OnlineId(target_type=match["target_type"], slug=match["slug"])
    return None



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Mbed Program abstraction layer."""
import logging

from pathlib import Path
from typing import List, Dict
from urllib.parse import urlparse

from mbed_tools.project.exceptions import ProgramNotFound, ExistingProgram, MbedOSNotFound
from mbed_tools.project._internal import git_utils
from mbed_tools.project._internal.project_data import (
    MbedProgramFiles,
    MbedOS,
    MBED_OS_REFERENCE_FILE_NAME,
    MBED_OS_DIR_NAME,
)
from mbed_tools.project._internal.libraries import LibraryReferences, MbedLibReference

logger = logging.getLogger(__name__)


class MbedProgram:
    """Represents an Mbed program.

    An `MbedProgram` consists of:
        * A copy of, or reference to, `MbedOS`
        * A set of `MbedProgramFiles`
        * A collection of references to external libraries, defined in .lib files located in the program source tree
    """

    def __init__(self, program_files: MbedProgramFiles, mbed_os: MbedOS) -> None:
        """Initialise the program attributes.

        Args:
            program_files: Object holding paths to a set of files that define an Mbed program.
            mbed_os: An instance of `MbedOS` holding paths to locations in the local copy of the Mbed OS source.
        """
        self.files = program_files
        self.root = self.files.mbed_os_ref.parent
        self.mbed_os = mbed_os
        self.lib_references = LibraryReferences(root=self.root, ignore_paths=[self.mbed_os.root.name])

    @classmethod
    def from_url(cls, url: str, dst_path: Path) -> "MbedProgram":
        """Fetch an Mbed program from a remote URL.

        Args:
            url: URL of the remote program repository.
            dst_path: Destination path for the cloned program.
        """
        logger.info(f"Cloning Mbed program from URL '{url}'.")
        git_utils.clone(url, dst_path)
        program_files = MbedProgramFiles.from_existing(dst_path)
        try:
            mbed_os = MbedOS.from_existing(dst_path / MBED_OS_DIR_NAME, False)
        except ValueError as mbed_err:
            raise MbedOSNotFound(f"{mbed_err}")

        return cls(program_files, mbed_os)

    @classmethod
    def from_new(cls, dir_path: Path) -> "MbedProgram":
        """Create an MbedProgram from an empty directory.

        Creates the directory if it doesn't exist.

        Args:
            dir_path: Directory in which to create the program.

        Raises:
            ExistingProgram: An existing program was found in the path.
        """
        if _tree_contains_program(dir_path):
            raise ExistingProgram(
                f"An existing Mbed program was found in the directory tree {dir_path}. It is not possible to nest Mbed "
                "programs. Please ensure there is no mbed-os.lib file in the cwd hierarchy."
            )

        logger.info(f"Creating Mbed program at path '{dir_path.resolve()}'")
        dir_path.mkdir(exist_ok=True)
        program_files = MbedProgramFiles.from_new(dir_path)
        logger.info(f"Creating git repository for the Mbed program '{dir_path}'")
        mbed_os = MbedOS.from_new(dir_path / MBED_OS_DIR_NAME)
        return cls(program_files, mbed_os)

    @classmethod
    def from_existing(cls, dir_path: Path, check_mbed_os: bool = True) -> "MbedProgram":
        """Create an MbedProgram from an existing program directory.

        Args:
            dir_path: Directory containing an Mbed program.
            check_mbed_os: If True causes an exception to be raised if the Mbed OS source directory does not
                           exist.

        Raises:
            ProgramNotFound: An existing program was not found in the path.
        """
        program_root = _find_program_root(dir_path)
        logger.info(f"Found existing Mbed program at path '{program_root}'")
        program = MbedProgramFiles.from_existing(program_root)

        try:
            mbed_os = MbedOS.from_existing(program_root / MBED_OS_DIR_NAME, check_mbed_os)
        except ValueError as mbed_os_err:
            raise MbedOSNotFound(
                f"Mbed OS was not found due to the following error: {mbed_os_err}"
                "\nYou may need to resolve the mbed-os.lib reference. You can do this by performing a `deploy`."
            )

        return cls(program, mbed_os)

    def resolve_libraries(self) -> None:
        """Resolve all external dependencies defined in .lib files."""
        self.lib_references.resolve()

    def deploy_libraries(self, force: bool = False) -> None:
        """Check out all resolved libraries to revisions specified in .lib files."""
        self.lib_references.deploy(force)

    def list_known_library_dependencies(self) -> List[MbedLibReference]:
        """Returns a list of all known library dependencies."""
        return sorted([lib for lib in self.lib_references.iter_all()])

    def has_unresolved_libraries(self) -> bool:
        """Checks if any unresolved library dependencies exist in the program tree."""
        return bool(list(self.lib_references.iter_unresolved()))


def parse_url(name_or_url: str) -> Dict[str, str]:
    """Create a valid github/armmbed url from a program name.

    Args:
        url: The URL, or a program name to turn into an URL.

    Returns:
        Dictionary containing the remote url and the destination path for the clone.
    """
    url_obj = urlparse(name_or_url)
    if url_obj.hostname:
        url = url_obj.geturl()
    else:
        url = f"https://github.com/armmbed/{url_obj.path}"
    # We need to create a valid directory name from the url path section.
    return {"url": url, "dst_path": url_obj.path.rsplit("/", maxsplit=1)[-1].replace("/", "")}


def _tree_contains_program(path: Path) -> bool:
    """Check if the current path or its ancestors contain an mbed-os.lib file.

    Args:
        path: The starting path for the search. The search walks up the tree from this path.

    Returns:
        `True` if an mbed-os.lib file is located between `path` and filesystem root.
        `False` if no mbed-os.lib file was found.
    """
    try:
        _find_program_root(path)
        return True
    except ProgramNotFound:
        return False


def _find_program_root(cwd: Path) -> Path:
    """Walk up the directory tree, looking for an mbed-os.lib file.

    Programs contain an mbed-os.lib file at the root of the source tree.

    Args:
        cwd: The directory path to search for a program.

    Raises:
        ProgramNotFound: No mbed-os.lib file found in the path.

    Returns:
        Path containing the mbed-os.lib file.
    """
    potential_root = cwd.absolute().resolve()
    while str(potential_root) != str(potential_root.anchor):
        logger.debug(f"Searching for mbed-os.lib file at path {potential_root}")
        root_file = potential_root / MBED_OS_REFERENCE_FILE_NAME
        if root_file.exists() and root_file.is_file():
            logger.debug(f"mbed-os.lib file found at {potential_root}")
            return potential_root

        potential_root = potential_root.parent

    logger.debug("No mbed-os.lib file found.")
    raise ProgramNotFound(
        f"No program found from {cwd.resolve()} to {cwd.resolve().anchor}. Please set the directory to a program "
        "directory containing an mbed-os.lib file. You can also set the directory to a program subdirectory if there "
        "is an mbed-os.lib file at the root of your program's directory tree."
    )



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a Disk partition/Logical disk relationship."""

from typing import NamedTuple

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor


class DiskToPartitionMsdnDefinition(NamedTuple):
    """Msdn definition of a disk partition - logical disk relationship.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/win32-logicaldisktopartition
    """

    EndingAddress: int
    StartingAddress: int
    Antecedent: str
    Dependent: str


class DiskPartitionLogicalDiskRelationship(ComponentDescriptor):
    """Disk partition as defined in Windows API.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/win32-logicaldisktopartition
    """

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(DiskToPartitionMsdnDefinition, win32_class_name="Win32_LogicalDiskToPartition")

    @property
    def component_id(self) -> str:
        """Returns the device id field."""
        return f"{self.get('Antecedent')}->{self.get('Dependent')}"

    @staticmethod
    def _parse_reference(ref: str) -> str:
        """Parses a Win32 to only retain the value."""
        return ref.replace("'", "").replace('"', "").split("=")[1]

    @property
    def logical_disk_id(self) -> str:
        """Gets the logical disk id."""
        return DiskPartitionLogicalDiskRelationship._parse_reference(self.get("Dependent"))

    @property
    def disk_partition_id(self) -> str:
        """Gets the disk partition id."""
        return DiskPartitionLogicalDiskRelationship._parse_reference(self.get("Antecedent"))



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib
import tempfile
from unittest import TestCase

from mbed_tools.build._internal.write_files import write_file
from mbed_tools.build.exceptions import InvalidExportOutputDirectory


class TestWriteFile(TestCase):
    def test_writes_content_to_file(self):
        with tempfile.TemporaryDirectory() as directory:
            content = "Some rendered content"
            export_path = pathlib.Path(directory, "output")
            file_name = "some_file.txt"

            write_file(export_path, file_name, content)

            created_file = pathlib.Path(export_path, pathlib.Path(export_path, file_name))
            self.assertEqual(created_file.read_text(), content)

    def test_output_dir_is_file(self):
        with tempfile.TemporaryDirectory() as directory:
            bad_export_dir = pathlib.Path(directory, "some_file.txt")
            bad_export_dir.touch()
            with self.assertRaises(InvalidExportOutputDirectory):
                write_file(bad_export_dir, "any_file.txt", "some contents")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Command to generate the application CMake configuration script used by the build system."""
import pathlib

import click

from mbed_tools.project import MbedProgram
from mbed_tools.build import generate_config


@click.command(
    help="Generate an Mbed OS config CMake file and write it to a .mbedbuild folder in the program directory."
)
@click.option(
    "-t",
    "--toolchain",
    type=click.Choice(["ARM", "GCC_ARM"], case_sensitive=False),
    required=True,
    help="The toolchain you are using to build your app.",
)
@click.option("-m", "--mbed-target", required=True, help="A build target for an Mbed-enabled device, eg. K64F")
@click.option(
    "-p",
    "--program-path",
    type=click.Path(),
    default=".",
    help="Path to local Mbed program. By default is the current working directory.",
)
def configure(toolchain: str, mbed_target: str, program_path: str) -> None:
    """Exports a mbed_config.cmake file to a .mbedbuild directory in the program root.

    The parameters set in the CMake file will be dependent on the combination of
    toolchain and Mbed target provided and these can then control which parts of
    Mbed OS are included in the build.

    This command will create the .mbedbuild directory at the program root if it doesn't
    exist.

    Args:
        toolchain: the toolchain you are using (eg. GCC_ARM, ARM)
        mbed_target: the target you are building for (eg. K64F)
        program_path: the path to the local Mbed program
    """
    program = MbedProgram.from_existing(pathlib.Path(program_path))
    mbed_target = mbed_target.upper()
    output_path = generate_config(mbed_target, toolchain, program)
    click.echo(f"mbed_config.cmake has been generated and written to '{str(output_path.resolve())}'")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Writes out files to specified locations."""
import pathlib
from mbed_tools.build.exceptions import InvalidExportOutputDirectory


def write_file(output_directory: pathlib.Path, file_name: str, file_contents: str) -> None:
    """Writes out a file to a directory.

    If the intermediate directories to the output directory don't exist,
    this function will create them.

    This function will overwrite any existing file of the same name in the
    output directory.

    Raises:
        InvalidExportOutputDirectory: it's not possible to export to the output directory provided
    """
    if output_directory.is_file():
        raise InvalidExportOutputDirectory("Output directory cannot be a path to a file.")

    output_directory.mkdir(parents=True, exist_ok=True)
    output_file = output_directory.joinpath(file_name)
    output_file.write_text(file_contents)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Find files in MbedOS program directory."""
from pathlib import Path
import fnmatch
from typing import Callable, Iterable, Optional, List, Tuple


def find_files(filename: str, directory: Path) -> List[Path]:
    """Proxy to `_find_files`, which applies legacy filtering rules."""
    # Temporary workaround, which replicates hardcoded ignore rules from old tools.
    # Legacy list of ignored directories is longer, however "TESTS" and
    # "TEST_APPS" were the only ones that actually exist in the MbedOS source.
    # Ideally, this should be solved by putting an `.mbedignore` file in the root of MbedOS repo,
    # similarly to what the code below pretends is happening.
    legacy_ignore = MbedignoreFilter(("*/TESTS", "*/TEST_APPS"))
    return _find_files(filename, directory, [legacy_ignore])


def _find_files(filename: str, directory: Path, filters: Optional[List[Callable]] = None) -> List[Path]:
    """Recursively find files by name under a given directory.

    This function automatically applies rules from .mbedignore files found during traversal.

    It is important to realise that applied filters are "greedy". The moment a directory is filtered out,
    its children won't be traversed.

    Args:
        filename: Name of the file to look for.
        directory: Location where search starts.
        filters: Optional list of exclude filters to apply.
    """
    if filters is None:
        filters = []

    result: List[Path] = []

    # Directories and files to process
    children = list(directory.iterdir())

    # If .mbedignore is one of the children, we need to add it to filter list,
    # as it might contain rules for currently processed directory, as well as its descendants.
    mbedignore = Path(directory, ".mbedignore")
    if mbedignore in children:
        filters = filters + [MbedignoreFilter.from_file(mbedignore)]

    # Remove files and directories that don't match current set of filters
    filtered_children = filter_files(children, filters)

    for child in filtered_children:
        if child.is_dir():
            # If processed child is a directory, recurse with current set of filters
            result += _find_files(filename, child, filters)

        if child.is_file() and child.name == filename:
            # We've got a match
            result.append(child)

    return result


def filter_files(files: Iterable[Path], filters: Iterable[Callable]) -> Iterable[Path]:
    """Filter given paths to files using filter callables."""
    return [file for file in files if all(f(file) for f in filters)]


class LabelFilter:
    """Filter out given paths using path labelling rules.

    If a path is labelled with given type, but contains label value which is
    not allowed, it will be filtered out.

    An example of labelled path is "/mbed-os/rtos/source/TARGET_CORTEX/mbed_lib.json",
    where label type is "TARGET" and label value is "CORTEX".

    For example, given a label type "FEATURE" and allowed values ["FOO"]:
    - "/path/FEATURE_FOO/somefile.txt" will not be filtered out
    - "/path/FEATURE_BAZ/somefile.txt" will be filtered out
    - "/path/FEATURE_FOO/FEATURE_BAR/somefile.txt" will be filtered out
    """

    def __init__(self, label_type: str, allowed_label_values: Iterable[str]):
        """Initialise the filter attributes.

        Args:
            label_type: Type of the label to filter with. In filtered paths, it prefixes the value.
            allowed_label_values: Values which are allowed for the given label type.
        """
        self._label_type = label_type
        self._allowed_labels = set(f"{label_type}_{label_value}" for label_value in allowed_label_values)

    def __call__(self, path: Path) -> bool:
        """Return True if given path contains only allowed labels - should not be filtered out."""
        labels = set(part for part in path.parts if self._label_type in part)
        return labels.issubset(self._allowed_labels)


class MbedignoreFilter:
    """Filter out given paths based on rules found in .mbedignore files.

    Patterns in .mbedignore use unix shell-style wildcards (fnmatch). It means
    that functionality, although similar is different to that found in
    .gitignore and friends.
    """

    def __init__(self, patterns: Tuple[str, ...]):
        """Initialise the filter attributes.

        Args:
            patterns: List of patterns from .mbedignore to filter against.
        """
        self._patterns = patterns

    def __call__(self, path: Path) -> bool:
        """Return True if given path doesn't match .mbedignore patterns - should not be filtered out."""
        stringified = str(path)
        return not any(fnmatch.fnmatch(stringified, pattern) for pattern in self._patterns)

    @classmethod
    def from_file(cls, mbedignore_path: Path) -> "MbedignoreFilter":
        """Return new instance with patterns read from .mbedignore file.

        Constructed patterns are rooted in the directory of .mbedignore file.
        """
        lines = mbedignore_path.read_text().splitlines()
        pattern_lines = (line for line in lines if line.strip() and not line.startswith("#"))
        ignore_root = mbedignore_path.parent
        patterns = tuple(str(ignore_root.joinpath(pattern)) for pattern in pattern_lines)
        return cls(patterns)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for `mbed_tools.targets.target_attributes`."""
import pathlib
import tempfile
from unittest import TestCase, mock

from mbed_tools.targets._internal.exceptions import TargetsJsonConfigurationError
from mbed_tools.targets._internal.target_attributes import (
    ParsingTargetsJSONError,
    TargetNotFoundError,
    get_target_attributes,
    _read_json_file,
    _extract_target_attributes,
    _extract_core_labels,
    _apply_config_overrides,
)


class TestExtractTargetAttributes(TestCase):
    def test_no_target_found(self):
        all_targets_data = {
            "Target_1": "some attributes",
            "Target_2": "some more attributes",
        }
        with self.assertRaises(TargetNotFoundError):
            _extract_target_attributes(all_targets_data, "Unlisted_Target")

    def test_target_found(self):
        target_attributes = {"attribute1": "something"}

        all_targets_data = {
            "Target_1": target_attributes,
            "Target_2": "some more attributes",
        }
        # When not explicitly included public is assumed to be True
        self.assertEqual(_extract_target_attributes(all_targets_data, "Target_1"), target_attributes)

    def test_target_public(self):
        all_targets_data = {
            "Target_1": {"attribute1": "something", "public": True},
            "Target_2": "some more attributes",
        }
        # The public attribute affects visibility but is removed from result
        self.assertEqual(_extract_target_attributes(all_targets_data, "Target_1"), {"attribute1": "something"})

    def test_target_private(self):
        all_targets_data = {
            "Target_1": {"attribute1": "something", "public": False},
            "Target_2": "some more attributes",
        }
        with self.assertRaises(TargetNotFoundError):
            _extract_target_attributes(all_targets_data, "Target_1"),


class TestReadTargetsJSON(TestCase):
    def test_valid_path(self):
        contents = """{
            "Target_Name": {
                "attribute_1": []
            }
        }"""
        with tempfile.TemporaryDirectory() as directory:
            json_file = pathlib.Path(directory, "targets.json")
            json_file.write_text(contents)
            result = _read_json_file(json_file)

            self.assertEqual(type(result), dict)

    def test_invalid_path(self):
        json_file = pathlib.Path("i_dont_exist")

        with self.assertRaises(FileNotFoundError):
            _read_json_file(json_file)

    def test_malformed_json(self):
        contents = """{
            "Target_Name": {
                []
            }
        }"""
        with tempfile.TemporaryDirectory() as directory:
            json_file = pathlib.Path(directory, "targets.json")
            json_file.write_text(contents)

            with self.assertRaises(ParsingTargetsJSONError):
                _read_json_file(json_file)


class TestGetTargetAttributes(TestCase):
    @mock.patch("mbed_tools.targets._internal.target_attributes._read_json_file")
    @mock.patch("mbed_tools.targets._internal.target_attributes._extract_target_attributes")
    @mock.patch("mbed_tools.targets._internal.target_attributes.get_labels_for_target")
    @mock.patch("mbed_tools.targets._internal.target_attributes._extract_core_labels")
    def test_gets_attributes_for_target(
        self, extract_core_labels, get_labels_for_target, extract_target_attributes, read_json_file
    ):
        targets_json_path = pathlib.Path("mbed-os/targets/targets.json")
        target_name = "My_Target"
        build_attributes = {"attribute": "value"}
        extract_target_attributes.return_value = build_attributes

        result = get_target_attributes(targets_json_path, target_name)

        read_json_file.assert_called_once_with(targets_json_path)
        extract_target_attributes.assert_called_once_with(read_json_file.return_value, target_name)
        get_labels_for_target.assert_called_once_with(read_json_file.return_value, target_name)
        extract_core_labels.assert_called_once_with(build_attributes.get("core", None))
        self.assertEqual(result, extract_target_attributes.return_value)


class TestExtractCoreLabels(TestCase):
    @mock.patch("mbed_tools.targets._internal.target_attributes._read_json_file")
    def test_extract_core(self, read_json_file):
        core_labels = ["FOO", "BAR"]
        metadata = {"CORE_LABELS": {"core_name": core_labels}}
        read_json_file.return_value = metadata
        target_core = "core_name"

        result = _extract_core_labels(target_core)

        self.assertEqual(result, set(core_labels))

    def test_no_core(self):
        result = _extract_core_labels(None)
        self.assertEqual(result, set())

    @mock.patch("mbed_tools.targets._internal.target_attributes._read_json_file")
    def test_no_labels(self, read_json_file):
        metadata = {"CORE_LABELS": {"not_the_same_core": []}}
        read_json_file.return_value = metadata

        result = _extract_core_labels("core_name")

        self.assertEqual(result, set())


class TestApplyConfigOverrides(TestCase):
    def test_applies_overrides(self):
        config = {"foo": {"help": "Do a foo", "value": 0}}
        overrides = {"foo": 9}
        expected_result = {"foo": {"help": "Do a foo", "value": 9}}

        self.assertEqual(expected_result, _apply_config_overrides(config, overrides))

    def test_applies_no_overrides(self):
        config = {"foo": {"help": "Do a foo", "value": 0}}
        overrides = {}

        self.assertEqual(config, _apply_config_overrides(config, overrides))

    def test_overriding_non_existing_config(self):
        config = {"foo": {"help": "Do a foo", "value": 0}}
        overrides = {"bar": 9}
        with self.assertRaises(TargetsJsonConfigurationError):
            _apply_config_overrides(config, overrides)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Ability to parse bootloader overrides from Sources."""
from dataclasses import dataclass, field, fields
from typing import Any, Iterable, Optional

from mbed_tools.build._internal.config.source import Source


@dataclass
class BootloaderOverride:
    """Representation of a bootloader override."""

    value: Any
    name: str
    set_by: str


@dataclass
class BootloaderOverrides:
    """Representation of cumulative attributes assembled during Source parsing."""

    # delivery overrides
    deliver_to_target: Optional[BootloaderOverride] = field(default=None)
    deliver_artifacts: Optional[BootloaderOverride] = field(default=None)
    delivery_dir: Optional[BootloaderOverride] = field(default=None)

    # rom overrides managed BL
    bootloader_img: Optional[BootloaderOverride] = field(default=None)
    restrict_size: Optional[BootloaderOverride] = field(default=None)
    header_format: Optional[BootloaderOverride] = field(default=None)
    header_offset: Optional[BootloaderOverride] = field(default=None)
    app_offset: Optional[BootloaderOverride] = field(default=None)
    # rom overrides unmanaged BL
    mbed_app_start: Optional[BootloaderOverride] = field(default=None)
    mbed_app_size: Optional[BootloaderOverride] = field(default=None)
    # rom overrides managed/unmanaged BL
    mbed_rom_start: Optional[BootloaderOverride] = field(default=None)
    mbed_rom_size: Optional[BootloaderOverride] = field(default=None)

    # ram overrides
    mbed_ram_start: Optional[BootloaderOverride] = field(default=None)
    mbed_ram_size: Optional[BootloaderOverride] = field(default=None)

    @classmethod
    def from_sources(cls, sources: Iterable[Source]) -> "BootloaderOverrides":
        """Interrogate each Source in turn to create BootloaderOverrides."""
        data = BootloaderOverrides()
        for source in sources:
            for key, value in source.overrides.items():
                if key in BOOTLOADER_OVERRIDE_KEYS_IN_SOURCE:
                    key_without_prefix = key[len("target.") :]
                    setattr(
                        data,
                        key_without_prefix,
                        BootloaderOverride(name=key_without_prefix, value=value, set_by=source.human_name),
                    )
        return data


_BOOTLOADER_OVERRIDE_FIELDS = [f.name for f in fields(BootloaderOverrides)]
BOOTLOADER_OVERRIDE_KEYS_IN_SOURCE = [f"target.{f}" for f in _BOOTLOADER_OVERRIDE_FIELDS]



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from unittest import TestCase, mock

from mbed_tools.project import initialise_project, clone_project, deploy_project, get_known_libs


@mock.patch("mbed_tools.project.project.MbedProgram", autospec=True)
class TestInitialiseProject(TestCase):
    def test_fetches_mbed_os_when_create_only_is_false(self, mock_program):
        path = pathlib.Path()
        initialise_project(path, create_only=False)

        mock_program.from_new.assert_called_once_with(path)
        mock_program.from_new.return_value.resolve_libraries.assert_called_once()

    def test_skips_mbed_os_when_create_only_is_true(self, mock_program):
        path = pathlib.Path()
        initialise_project(path, create_only=True)

        mock_program.from_new.assert_called_once_with(path)
        mock_program.from_new.return_value.resolve_libraries.assert_not_called()


@mock.patch("mbed_tools.project.project.MbedProgram", autospec=True)
class TestCloneProject(TestCase):
    def test_clones_from_remote(self, mock_program):
        url = "https://git.com/gitorg/repo"
        clone_project(url, recursive=False)

        mock_program.from_url.assert_called_once_with(url, pathlib.Path(url.rsplit("/", maxsplit=1)[-1]))

    def test_resolves_libs_when_recursive_is_true(self, mock_program):
        url = "https://git.com/gitorg/repo"
        clone_project(url, recursive=True)

        mock_program.from_url.assert_called_once_with(url, pathlib.Path(url.rsplit("/", maxsplit=1)[-1]))
        mock_program.from_url.return_value.resolve_libraries.assert_called_once()


@mock.patch("mbed_tools.project.project.MbedProgram", autospec=True)
class TestDeployProject(TestCase):
    def test_checks_out_libraries(self, mock_program):
        path = pathlib.Path("somewhere")
        deploy_project(path, force=False)

        mock_program.from_existing.assert_called_once_with(path, False)
        mock_program.from_existing.return_value.deploy_libraries.assert_called_once_with(force=False)

    def test_resolves_libs_if_unresolved_detected(self, mock_program):
        path = pathlib.Path("somewhere")
        deploy_project(path)

        mock_program.from_existing.return_value.resolve_libraries.assert_called_once()


@mock.patch("mbed_tools.project.project.MbedProgram", autospec=True)
class TestPrintLibs(TestCase):
    def test_list_libraries_called(self, mock_program):
        path = pathlib.Path("somewhere")
        get_known_libs(path)

        mock_program.from_existing.assert_called_once_with(path, False)
        mock_program.from_existing.return_value.list_known_library_dependencies.assert_called()



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a generic Win32 component."""
import logging
from abc import ABC, abstractmethod
from typing import List, Any, Generator, Optional, NamedTuple, cast

import pythoncom
import win32com.client

from mbed_tools.devices._internal.windows.component_descriptor_utils import (
    UNKNOWN_VALUE,
    is_undefined_data_object,
)

NAMED_TUPLE_FIELDS_ATTRIBUTE = "_fields"

logger = logging.getLogger(__name__)


class ComponentDescriptor(ABC):
    """Win32 component descriptor."""

    def __init__(self, win32_definition: type, win32_class_name: str, win32_filter: Optional[str] = None):
        """Initialiser.

        Args:
            win32_definition: definition of the Windows component as defined in MSDN.
            win32_class_name: Win32 class name of the component
            win32_filter: Any extra filter to apply such as a subcategory of a Win32 class.
        """
        self._win32_definition = win32_definition
        self._win32_class_name = win32_class_name
        self._win32_filter = win32_filter

    def set_data_values(self, fields_values: dict) -> None:
        """Sets fields values based on what is defined in dictionary."""
        for k, v in fields_values.items():
            setattr(self, k, v)

    @property
    def win32_definition(self) -> type:
        """Gets descriptor definition."""
        return self._win32_definition

    @property
    def win32_class_name(self) -> str:
        """Returns the name of the Win32 Class."""
        return self._win32_class_name

    @property
    def field_names(self) -> List[str]:
        """Returns the names of all the fields of the descriptor."""
        return [k for k in getattr(self._win32_definition, NAMED_TUPLE_FIELDS_ATTRIBUTE)]

    @property
    @abstractmethod
    def component_id(self) -> str:
        """Returns the device ID field."""

    @property
    def win32_filter(self) -> Optional[str]:
        """Filter applied on a Win32 category.

        For instance, the current component can be a subclass/subcategory of a component exposed by Win32.
        """
        return self._win32_filter

    def to_tuple(self) -> NamedTuple:
        """Translates into named tuple."""
        return cast(NamedTuple, self.win32_definition(**{k: self.get(k) for k in self.field_names}))

    @property
    def is_undefined(self) -> bool:
        """Determines whether the structure is undefined or not i.e. none of the fields are actually defined."""
        return is_undefined_data_object(self.to_tuple())

    def get(self, field_name: str) -> Any:
        """Gets the field value."""
        try:
            return getattr(self, field_name)
        except AttributeError as e:
            logger.debug(f"Attribute [{field_name}] is undefined on this instance {self}: {e}")
            return UNKNOWN_VALUE

    def __str__(self) -> str:
        """String representation."""
        values = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        return f"{self.__class__.__name__}({values})"


class Win32Wrapper:
    """Wraps win32 objects and methods in order to simplify their use."""

    def __init__(self) -> None:
        """Wrapper initialisation."""
        # Setting pyWin32 so that it can be used across multiple threads
        # See:
        # https://stackoverflow.com/questions/37258257/why-does-this-script-not-work-with-threading-python
        # https://gist.github.com/vlasenkov/f8fe40d5b2d9e43fd46ad8363067acce
        # https://stackoverflow.com/questions/26764978/using-win32com-with-multithreading/27966218#27966218
        pythoncom.CoInitialize()
        self.wmi = win32com.client.GetObject("winmgmts:")

    def _read_cdispatch_fields(self, win32_element: Any, element_fields_list: List[str]) -> dict:
        """Reads all the fields from a cdispatch object returned by pywin32."""
        if not win32_element:
            return dict()
        return {k: self._read_cdispatch_field(win32_element, k) for k in element_fields_list}

    def _read_cdispatch_field(self, win32_element: Any, key: str) -> Any:
        """Reads a specific field on a cdispatch object."""
        try:
            return getattr(win32_element, key)
        except AttributeError as e:
            logger.debug(f"Failed getting an attribute value: {e}")
            return UNKNOWN_VALUE

    def map_element(self, win32_element: Any, to_cls: type) -> ComponentDescriptor:
        """Maps a win32 element into an element of the class `to_cls`."""
        instance = to_cls()
        instance.set_data_values(self._read_cdispatch_fields(win32_element, instance.field_names))
        return cast(ComponentDescriptor, instance)

    def _get_list_iterator(self, win32_class_name: str, list_filter: Optional[str]) -> Generator[Any, None, None]:
        if list_filter:
            query = f"Select * from {win32_class_name} where {list_filter}"
            return self.wmi.ExecQuery(query)  # type: ignore
        return self.wmi.InstancesOf(win32_class_name)  # type: ignore

    def element_generator(
        self, to_cls: type, win32_class_name: str, list_filter: Optional[str]
    ) -> Generator["ComponentDescriptor", None, None]:
        """Gets a generator over all elements currently registered in the system."""
        for element in self._get_list_iterator(win32_class_name, list_filter):
            yield self.map_element(win32_element=element, to_cls=to_cls)


class ComponentDescriptorWrapper:
    """Wraps a component descriptor."""

    def __init__(self, cls: type):
        """initialiser."""
        self._cls = cls
        self._win32_wrapper = Win32Wrapper()

    def element_generator(self) -> Generator["ComponentDescriptor", None, None]:
        """Gets a generator over all elements currently registered in the system."""
        instance = self._cls()
        return self._win32_wrapper.element_generator(self._cls, instance.win32_class_name, instance.win32_filter)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from dataclasses import fields
from unittest import TestCase

from mbed_tools.build._internal.config.cumulative_data import CumulativeData
from tests.build._internal.config.factories import SourceFactory


class TestCumulativeDataFromSources(TestCase):
    def test_assembles_metadata_from_sources(self):
        for field in fields(CumulativeData):
            with self.subTest(f"Assemble {field.name}"):
                source_a = SourceFactory(overrides={f"target.{field.name}": ["FOO"]})
                source_b = SourceFactory(overrides={f"target.{field.name}_add": ["BAR", "BAZ"]})
                source_c = SourceFactory(overrides={f"target.{field.name}_remove": ["BAR"]})

                config_cumulative_data = CumulativeData.from_sources([source_a, source_b, source_c])

                self.assertEqual(getattr(config_cumulative_data, field.name), {"FOO", "BAZ"})



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase, mock

from tests.devices.markers import windows_only, darwin_only, linux_only
from mbed_tools.devices._internal.base_detector import DeviceDetector
from mbed_tools.devices._internal.detect_candidate_devices import (
    detect_candidate_devices,
    _get_detector_for_current_os,
)


class TestDetectCandidateDevices(TestCase):
    @mock.patch("mbed_tools.devices._internal.detect_candidate_devices._get_detector_for_current_os")
    def test_returns_candidates_using_os_specific_detector(self, _get_detector_for_current_os):
        detector = mock.Mock(spec_set=DeviceDetector)
        _get_detector_for_current_os.return_value = detector
        self.assertEqual(detect_candidate_devices(), detector.find_candidates.return_value)


class TestGetDetectorForCurrentOS(TestCase):
    @windows_only
    def test_windows_uses_correct_module(self):
        from mbed_tools.devices._internal.windows.device_detector import WindowsDeviceDetector

        self.assertIsInstance(_get_detector_for_current_os(), WindowsDeviceDetector)

    @darwin_only
    def test_darwin_uses_correct_module(self):
        from mbed_tools.devices._internal.darwin.device_detector import DarwinDeviceDetector

        self.assertIsInstance(_get_detector_for_current_os(), DarwinDeviceDetector)

    @linux_only
    def test_linux_uses_correct_module(self):
        from mbed_tools.devices._internal.linux.device_detector import LinuxDeviceDetector

        self.assertIsInstance(_get_detector_for_current_os(), LinuxDeviceDetector)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Functions for parsing the inheritance for accumulating attributes.

Accumulating attributes are both defined and can be added to and removed from further down the hierarchy.
The hierarchy is also slightly different to the other fields as it is determined as 'breadth-first' in
multiple inheritance, so targets at a lower level will always take precedence over targets at a higher level.
"""
import itertools
from collections import deque
from typing import Dict, List, Any, Deque

ACCUMULATING_ATTRIBUTES = ("extra_labels", "macros", "device_has", "features", "components")
MODIFIERS = ("add", "remove")
ALL_ACCUMULATING_ATTRIBUTES = ACCUMULATING_ATTRIBUTES + tuple(
    f"{attr}_{suffix}" for attr, suffix in itertools.product(ACCUMULATING_ATTRIBUTES, MODIFIERS)
)


def get_accumulating_attributes_for_target(all_targets_data: Dict[str, Any], target_name: str) -> Dict[str, Any]:
    """Parses the data for all targets and returns the accumulating attributes for the specified target.

    Args:
        all_targets_data: a dictionary representation of the contents of targets.json
        target_name: the name of the target to find the attributes of

    Returns:
        A dictionary containing all the accumulating attributes for the chosen target
    """
    accumulating_order = _targets_accumulate_hierarchy(all_targets_data, target_name)
    return _determine_accumulated_attributes(accumulating_order)


def _targets_accumulate_hierarchy(all_targets_data: Dict[str, Any], target_name: str) -> List[dict]:
    """List all ancestors of a target in order of accumulation inheritance (breadth-first).

    Using a breadth-first traverse of the inheritance tree, return a list of targets in the
    order of inheritance, starting with the target itself and finishing with its highest ancestor.

    Eg. An inheritance tree diagram for target "A" below

    D     E
    |     |
    B     C
    |_____|
       |
       A

    Would give us and inheritance order of [A, B, C, D, E]

    Args:
        all_targets_data: a dictionary representation of all the data in a targets.json file
        target_name: the name of the target we want to calculate the attributes for

    Returns:
        A list of dicts representing each target in the hierarchy.
    """
    targets_in_order: List[dict] = []

    still_to_visit: Deque[dict] = deque()
    still_to_visit.appendleft(all_targets_data[target_name])

    while still_to_visit:
        current_target = still_to_visit.popleft()
        targets_in_order.append(current_target)
        for parent_target in current_target.get("inherits", []):
            still_to_visit.append(all_targets_data[parent_target])

    return targets_in_order


def _add_attribute_element(
    accumulator: Dict[str, Any], attribute_name: str, elements_to_add: List[Any]
) -> Dict[str, Any]:
    """Adds an attribute element to an attribute.

    Args:
        accumulator: a store of attributes to be updated
        attribute_name: name of the attribute to update
        elements_to_add: element to add to the attribute list

    Returns:
        The accumulator object with the new elements added
    """
    for element in elements_to_add:
        accumulator[attribute_name].append(element)
    return accumulator


def _element_matches(element_to_remove: str, element_to_check: str) -> bool:
    """Checks if an element meets the criteria to be removed from list.

    Some attribute elements (eg. macros) can be defined with a number value
    eg. MACRO_SOMETHING=5. If we are then instructed to remove
    MACRO_SOMETHING then this element needs to be recognised and removed
    in addition to exact matches.

    Args:
        element_to_remove: the element as taken from list to be removed from an attribute
        element_to_check: an element that currently makes up part of an attribute definition

    Returns:
        A boolean reflecting whether the element is a match and should be removed
    """
    return element_to_check == element_to_remove or element_to_check.startswith(f"{element_to_remove}=")


def _remove_attribute_element(
    accumulator: Dict[str, Any], attribute_name: str, elements_to_remove: List[Any]
) -> Dict[str, Any]:
    """Removes an attribute element from an attribute.

    Args:
        accumulator: a store of attributes to be updated
        attribute_name: name of the attribute to update
        elements_to_remove: element to remove from the attribute list

    Returns:
        The accumulator object with the desired elements removed
    """
    existing_elements = accumulator[attribute_name]
    combinations_to_check = itertools.product(existing_elements, elements_to_remove)
    checked_elements_to_remove = [
        existing_element
        for existing_element, element in combinations_to_check
        if _element_matches(element, existing_element)
    ]

    for element in checked_elements_to_remove:
        accumulator[attribute_name].remove(element)
    return accumulator


def _calculate_attribute_elements(
    attribute_name: str, starting_state: Dict[str, Any], applicable_accumulation_order: List[dict]
) -> Dict[str, Any]:
    """Adds and removes elements for an attribute based on the definitions encountered in the hierarchy.

    This is done via modifying attributes eg. "extra_labels" can be modified by
    "extra_labels_add" and "extra_labels_remove".

    Args:
        attribute_name: name of the attribute to update
        starting_state: the list of elements that defines the starting state of the attribute
        applicable_accumulation_order: the targets in the inheritance tree that can contain add and remove modifiers

    Returns:
        The elements calculated for an attribute according to the target's inheritance
    """
    accumulator = starting_state
    for target in reversed(applicable_accumulation_order):

        add_modifier = f"{attribute_name}_add"
        if add_modifier in target:
            to_add = target[add_modifier]
            _add_attribute_element(accumulator, attribute_name, to_add)

        remove_modifier = f"{attribute_name}_remove"
        if remove_modifier in target:
            to_remove = target[remove_modifier]
            _remove_attribute_element(accumulator, attribute_name, to_remove)

    return accumulator


def _calculate_attribute_for_target(
    attribute_name: str, target: Dict[str, Any], targets_in_order: List[dict]
) -> Dict[str, Any]:
    """Finds a single accumulated attribute for a target from its list of ancestors.

    Args:
        attribute_name: the name of the accumulating attribute
        target: the target we are collecting data for
        targets_in_order: the full inheritance hierarchy for a target

    Returns:
        A dictionary representation of a single accumulating attribute for that target
    """
    starting_state = {attribute_name: target[attribute_name]}
    # Reduces the order list to only the targets in the hierarchy between the starting state and the target itself
    applicable_accumulation_order = targets_in_order[: targets_in_order.index(target)]
    return _calculate_attribute_elements(attribute_name, starting_state, applicable_accumulation_order)


def _find_nearest_defined_attribute(targets_in_order: List[dict], attribute_name: str) -> Dict[str, Any]:
    """Returns the definition of a particular attribute first encountered in the accumulation order.

    Args:
        targets_in_order: the inheritance order for the target, from the target itself to its highest ancestor
        attribute_name: the attribute to search for

    Returns:
        A dictionary containing the definition of the requested attribute
    """
    for target in targets_in_order:
        if attribute_name in target:
            return _calculate_attribute_for_target(attribute_name, target, targets_in_order)
    return {}


def _determine_accumulated_attributes(targets_in_order: List[dict]) -> Dict[str, Any]:
    """Finds all the accumulated attributes for a target from its list of ancestors.

    Iterates through the order of inheritance (accumulation order) to find the nearest definition
    of an attribute, then retraces backwards calculating additions and deletions that modify it.

    Args:
        targets_in_order: the inheritance order for the target, from the target itself to its highest ancestor

    Returns:
        A dictionary containing all the accumulating attributes for a target
    """
    accumulated_attributes = {}

    for attribute_name in ACCUMULATING_ATTRIBUTES:
        accumulated_attributes.update(_find_nearest_defined_attribute(targets_in_order, attribute_name))
    return accumulated_attributes



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import plistlib
from unittest import TestCase, mock

from mbed_tools.devices._internal.darwin.ioreg import get_data, get_io_dialin_device


@mock.patch("mbed_tools.devices._internal.darwin.ioreg.subprocess.check_output")
class TestGetData(TestCase):
    def test_returns_data_from_ioreg_call(self, check_output):
        check_output.return_value = b"""<plist version="1.0">
            <array>
                <string>foo</string>
            </array>
        </plist>"""

        self.assertEqual(get_data("some device"), ["foo"])
        check_output.assert_called_once_with(["ioreg", "-a", "-r", "-n", "some device", "-l"])

    def test_handles_corrupt_data_gracefully(self, check_output):
        check_output.return_value = b"""<plist version="1.0">
            <string>S\xc3\xbfn\x06P\xc2\x87TT%A\t\xc2\x87</string>
        </plist>"""

        self.assertEqual(get_data("doesn't matter"), [])


class TestGetIoDialinDevice(TestCase):
    @mock.patch("mbed_tools.devices._internal.darwin.ioreg.get_data")
    def test_identifies_nested_io_dialin_device_in_given_ioreg_data(self, get_data):
        plist = b"""<plist version="1.0">
            <array>
              <dict>
                    <key>IORegistryEntryChildren</key>
                <array/>
              </dict>
              <dict>
                <key>IORegistryEntryChildren</key>
                <array>
                  <dict>
                    <key>IORegistryEntryChildren</key>
                    <array>
                      <dict>
                        <key>IODialinDevice</key>
                        <string>/dev/tty.usbmodem1234</string>
                      </dict>
                    </array>
                  </dict>
                </array>
              </dict>
            </array>
        </plist>"""
        data = plistlib.loads(plist)
        get_data.return_value = data

        self.assertEqual(get_io_dialin_device("some_device"), "/dev/tty.usbmodem1234")
        get_data.assert_called_once_with("some_device")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import json
import pathlib
from click.testing import CliRunner
from mbed_tools.devices.device import ConnectedDevices
from mbed_tools.targets import Board
from tabulate import tabulate
from unittest import TestCase, mock

from mbed_tools.cli.list_connected_devices import (
    list_connected_devices,
    _build_tabular_output,
    _build_json_output,
    _get_build_targets,
    _sort_devices,
)
from mbed_tools.devices import Device


@mock.patch("mbed_tools.cli.list_connected_devices.get_connected_devices")
class TestListConnectedDevices(TestCase):
    def test_informs_when_no_devices_are_connected(self, get_connected_devices):
        get_connected_devices.return_value = ConnectedDevices()

        result = CliRunner().invoke(list_connected_devices)

        self.assertEqual(result.exit_code, 0)
        self.assertIn("No connected Mbed devices found.", result.output)

    @mock.patch("mbed_tools.cli.list_connected_devices._sort_devices")
    @mock.patch("mbed_tools.cli.list_connected_devices._build_tabular_output")
    def test_by_default_lists_devices_using_tabular_output(
        self, _build_tabular_output, _sort_devices, get_connected_devices
    ):
        identified_devices = [mock.Mock(spec_set=Device)]
        unidentified_devices = [mock.Mock(spec_set=Device)]
        get_connected_devices.return_value = ConnectedDevices(
            identified_devices=identified_devices, unidentified_devices=unidentified_devices
        )
        _build_tabular_output.return_value = "some output"

        result = CliRunner().invoke(list_connected_devices)

        self.assertEqual(result.exit_code, 0)
        self.assertIn(_build_tabular_output.return_value, result.output)
        _build_tabular_output.assert_called_once_with(_sort_devices.return_value)
        _sort_devices.assert_called_once_with(identified_devices)

    @mock.patch("mbed_tools.cli.list_connected_devices._sort_devices")
    @mock.patch("mbed_tools.cli.list_connected_devices._build_json_output")
    def test_given_json_flag_lists_devices_using_json_output(
        self, _build_json_output, _sort_devices, get_connected_devices
    ):
        identified_devices = [mock.Mock(spec_set=Device)]
        unidentified_devices = [mock.Mock(spec_set=Device)]
        get_connected_devices.return_value = ConnectedDevices(
            identified_devices=identified_devices, unidentified_devices=unidentified_devices
        )
        _build_json_output.return_value = "some output"

        result = CliRunner().invoke(list_connected_devices, "--format=json")

        self.assertEqual(result.exit_code, 0)
        self.assertIn(_build_json_output.return_value, result.output)
        _build_json_output.assert_called_once_with(_sort_devices.return_value)
        _sort_devices.assert_called_once_with(identified_devices)

    @mock.patch("mbed_tools.cli.list_connected_devices._sort_devices")
    @mock.patch("mbed_tools.cli.list_connected_devices._build_tabular_output")
    def test_given_show_all(self, _build_tabular_output, _sort_devices, get_connected_devices):
        identified_devices = [mock.Mock(spec_set=Device)]
        unidentified_devices = [mock.Mock(spec_set=Device)]
        get_connected_devices.return_value = ConnectedDevices(
            identified_devices=identified_devices, unidentified_devices=unidentified_devices
        )
        _build_tabular_output.return_value = "some output"

        result = CliRunner().invoke(list_connected_devices, "--show-all")

        self.assertEqual(result.exit_code, 0)
        self.assertIn(_build_tabular_output.return_value, result.output)
        _build_tabular_output.assert_called_once_with(_sort_devices.return_value)
        _sort_devices.assert_called_once_with(identified_devices + unidentified_devices)


class TestSortDevices(TestCase):
    def test_sorts_devices_by_board_name(self):
        device_1 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name="A"), serial_number="123"
        )
        device_2 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name="B"), serial_number="456"
        )
        device_3 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name="C"), serial_number="789"
        )

        result = _sort_devices([device_3, device_1, device_2])

        self.assertEqual(list(result), [device_1, device_2, device_3])

    def test_sorts_devices_by_serial_number(self):
        device_1 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name=""), serial_number="123"
        )
        device_2 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name=""), serial_number="456"
        )
        device_3 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name=""), serial_number="789"
        )

        result = _sort_devices([device_3, device_1, device_2])

        self.assertEqual(list(result), [device_1, device_2, device_3])

    def test_sorts_devices_by_board_name_then_serial_number(self):
        device_1 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name=""), serial_number="123"
        )
        device_2 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name="Mbed"), serial_number="456"
        )
        device_3 = mock.create_autospec(
            Device, mbed_board=mock.create_autospec(Board, board_name=""), serial_number="789"
        )

        result = _sort_devices([device_3, device_1, device_2])

        self.assertEqual(list(result), [device_1, device_3, device_2])


class TestBuildTableOutput(TestCase):
    def test_returns_tabularised_representation_of_devices(self):
        device = Device(
            mbed_board=mock.create_autospec(
                Board, board_name="board-name", build_variant=("S", "NS"), board_type="board-type",
            ),
            serial_number="serial-number",
            serial_port="serial-port",
            mount_points=[pathlib.Path("/Volumes/FOO"), pathlib.Path("/Volumes/BAR")],
        )

        output = _build_tabular_output([device])

        expected_output = tabulate(
            [
                [
                    device.mbed_board.board_name,
                    device.serial_number,
                    device.serial_port,
                    "\n".join(map(str, device.mount_points)),
                    "\n".join(_get_build_targets(device.mbed_board)),
                ]
            ],
            headers=["Board name", "Serial number", "Serial port", "Mount point(s)", "Build target(s)"],
        )
        self.assertEqual(output, expected_output)

    def test_displays_unknown_serial_port_value(self):
        device = Device(
            mbed_board=Board.from_offline_board_entry({}),
            serial_number="serial",
            serial_port=None,
            mount_points=[pathlib.Path("somepath")],
        )

        output = _build_tabular_output([device])

        expected_output = tabulate(
            [
                [
                    "<unknown>",
                    device.serial_number,
                    "<unknown>",
                    "\n".join(map(str, device.mount_points)),
                    "\n".join(_get_build_targets(device.mbed_board)),
                ]
            ],
            headers=["Board name", "Serial number", "Serial port", "Mount point(s)", "Build target(s)"],
        )
        self.assertEqual(output, expected_output)


class TestBuildJsonOutput(TestCase):
    def test_returns_json_representation_of_devices(self):
        board = mock.create_autospec(
            Board,
            product_code="0021",
            board_type="HAT-BOAT",
            board_name="HAT Boat",
            mbed_os_support=["0.2"],
            mbed_enabled=["potentially"],
            build_variant=("S", "NS"),
        )
        device = Device(
            mbed_board=board, serial_number="09887654", serial_port="COM1", mount_points=[pathlib.Path("somepath")],
        )

        output = _build_json_output([device])
        expected_output = json.dumps(
            [
                {
                    "serial_number": device.serial_number,
                    "serial_port": device.serial_port,
                    "mount_points": [str(m) for m in device.mount_points],
                    "mbed_board": {
                        "product_code": board.product_code,
                        "board_type": board.board_type,
                        "board_name": board.board_name,
                        "mbed_os_support": board.mbed_os_support,
                        "mbed_enabled": board.mbed_enabled,
                        "build_targets": _get_build_targets(board),
                    },
                }
            ],
            indent=4,
        )

        self.assertEqual(output, expected_output)

    def test_empty_values_keys_are_always_present(self):
        """Asserts that keys are present even if value is None."""
        device = Device(
            mbed_board=Board.from_offline_board_entry({}), serial_number="foo", serial_port=None, mount_points=[],
        )

        output = json.loads(_build_json_output([device]))

        self.assertIsNone(output[0]["serial_port"])


class TestGetBuildTargets(TestCase):
    def test_returns_base_target_and_all_variants(self):
        board = mock.create_autospec(Board, build_variant=("S", "NS"), board_type="FOO")

        self.assertEqual(_get_build_targets(board), ["FOO_S", "FOO_NS", "FOO"])



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""API for listing devices."""

from typing import Optional

from mbed_tools.targets.exceptions import MbedTargetsError
from mbed_tools.targets import Board

from mbed_tools.devices._internal.detect_candidate_devices import detect_candidate_devices
from mbed_tools.devices._internal.resolve_board import resolve_board
from mbed_tools.devices._internal.exceptions import NoBoardForCandidate

from mbed_tools.devices.device import ConnectedDevices
from mbed_tools.devices.exceptions import DeviceLookupFailed


def get_connected_devices() -> ConnectedDevices:
    """Returns Mbed Devices connected to host computer.

    Connected devices which have been identified as Mbed Boards and also connected devices which are potentially
    Mbed Boards (but not could not be identified in the database) are returned.

    Raises:
        DeviceLookupFailed: If there is a problem with the process of identifying Mbed Boards from connected devices.
    """
    connected_devices = ConnectedDevices()

    board: Optional["Board"]
    for candidate_device in detect_candidate_devices():
        try:
            board = resolve_board(candidate_device)
        except NoBoardForCandidate:
            board = None
        except MbedTargetsError as err:
            raise DeviceLookupFailed("A problem occurred when looking up board data for connected devices.") from err

        connected_devices.add_device(candidate_device, board)

    return connected_devices



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import json
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from mbed_tools.build._internal.config.assemble_build_config import (
    _assemble_config_from_sources_and_lib_files,
    assemble_config,
)
from mbed_tools.build._internal.config.config import Config
from mbed_tools.build._internal.find_files import find_files
from mbed_tools.build._internal.config.source import Source


def create_files(directory, files):
    created_files = []
    for file in files:
        path = Path(directory, file["path"])
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json.dumps(file["json_contents"]))
        created_files.append(path)
    return created_files


@mock.patch("mbed_tools.build._internal.config.assemble_build_config.Source", autospec=True)
@mock.patch("mbed_tools.build._internal.config.assemble_build_config.find_files", autospec=True)
@mock.patch(
    "mbed_tools.build._internal.config.assemble_build_config._assemble_config_from_sources_and_lib_files", autospec=True
)
class TestAssembleConfig(TestCase):
    def test_calls_collaborator_with_source_and_file_paths(
        self, _assemble_config_from_sources_and_lib_files, find_files, Source,
    ):
        mbed_target = dict()
        mbed_program_directory = Path("foo")
        app_config_file = mbed_program_directory / "mbed_app.json"

        subject = assemble_config(mbed_target, mbed_program_directory, app_config_file)

        self.assertEqual(subject, _assemble_config_from_sources_and_lib_files.return_value)
        _assemble_config_from_sources_and_lib_files.assert_called_once_with(
            mbed_target, find_files.return_value, app_config_file
        )
        find_files.assert_called_once_with("mbed_lib.json", mbed_program_directory)


class TestAssembleConfigFromSourcesAndLibFiles(TestCase):
    def test_assembles_config_using_all_relevant_files(self):
        target = {
            "config": {"foo": None},
            "macros": {},
            "labels": set("A"),
            "extra_labels": set(),
            "features": set("RED"),
            "components": set(),
        }
        mbed_lib_files = [
            {
                "path": Path("TARGET_A", "mbed_lib.json"),
                "json_contents": {
                    "name": "a",
                    "config": {"number": 123},
                    "target_overrides": {"*": {"target.features_add": ["RED"]}},
                },
            },
            {
                "path": Path("subdir", "FEATURE_RED", "mbed_lib.json"),
                "json_contents": {
                    "name": "red",
                    "config": {"bool": False},
                    "target_overrides": {
                        "A": {"bool": True, "target.features_add": ["BLUE"], "target.components_add": ["LEG"]}
                    },
                    "macros": ["RED_MACRO"],
                },
            },
            {
                "path": Path("COMPONENT_LEG", "mbed_lib.json"),
                "json_contents": {"name": "leg", "config": {"number-of-fingers": 5}, "macros": ["LEG_MACRO"]},
            },
        ]
        unused_mbed_lib_file = {
            "path": Path("subdir", "FEATURE_BROWN", "mbed_lib.json"),
            "json_contents": {
                "name": "brown",
                "target_overrides": {"*": {"red.bool": "DON'T USE ME"}},
                "macros": ["DONT_USE_THIS_MACRO"],
            },
        }
        mbed_app_file = {
            "path": Path("mbed_app.json"),
            "json_contents": {"target_overrides": {"*": {"target.foo": "bar"}}},
        }

        with TemporaryDirectory() as directory:
            created_mbed_lib_files = create_files(directory, mbed_lib_files)
            created_mbed_app_file = create_files(directory, [mbed_app_file])[0]
            create_files(directory, [unused_mbed_lib_file])

            subject = _assemble_config_from_sources_and_lib_files(
                target, find_files("mbed_lib.json", Path(directory)), created_mbed_app_file
            )

            mbed_lib_sources = [Source.from_mbed_lib(Path(directory, file), ["A"]) for file in created_mbed_lib_files]
            mbed_app_source = Source.from_mbed_app(created_mbed_app_file, ["A"])
            expected_config = Config.from_sources([Source.from_target(target)] + mbed_lib_sources + [mbed_app_source])

            self.assertEqual(subject, expected_config)

    def test_updates_target_labels_from_config(self):
        target = {
            "config": {"foo": None},
            "macros": {"NO"},
            "labels": set("A"),
            "extra_labels": set(),
            "features": set(),
            "components": set(),
        }
        mbed_lib_files = [
            {
                "path": Path("TARGET_A", "mbed_lib.json"),
                "json_contents": {"name": "a", "target_overrides": {"*": {"target.features_add": ["RED"]}}},
            },
            {
                "path": Path("subdir", "FEATURE_RED", "mbed_lib.json"),
                "json_contents": {
                    "name": "red",
                    "target_overrides": {
                        "A": {
                            "target.macros_remove": ["NO"],
                            "target.features_add": ["BLUE"],
                            "target.components_add": ["LEG"],
                        }
                    },
                    "macros": ["RED_MACRO"],
                },
            },
        ]
        mbed_app_file = {
            "path": Path("mbed_app.json"),
            "json_contents": {
                "target_overrides": {
                    "*": {
                        "target.foo": "bar",
                        "target.labels_add": ["PICKLE"],
                        "target.extra_labels_add": ["EXTRA_HOT"],
                        "target.macros_add": ["TICKER"],
                    }
                }
            },
        }

        with TemporaryDirectory() as directory:
            _ = create_files(directory, mbed_lib_files)
            created_mbed_app_file = create_files(directory, [mbed_app_file])[0]

            _ = _assemble_config_from_sources_and_lib_files(
                target, find_files("mbed_lib.json", Path(directory)), created_mbed_app_file
            )

            self.assertEqual(target["features"], {"RED", "BLUE"})
            self.assertEqual(target["components"], {"LEG"})
            self.assertEqual(target["extra_labels"], {"EXTRA_HOT"})
            self.assertEqual(target["labels"], {"A", "PICKLE"})
            self.assertEqual(target["macros"], {"TICKER"})



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for `mbed_tools.targets.mbed_target`."""

from unittest import TestCase

# Import from top level as this is the expected interface for users
from mbed_tools.targets import Board


class TestBoard(TestCase):
    """Tests for the class `Board`."""

    def test_offline_database_entry(self):
        """Given an entry from the offline database, a Board is generated with the correct information."""
        board = Board.from_offline_board_entry(
            {
                "mbed_os_support": ["Mbed OS 5.15"],
                "mbed_enabled": ["Basic"],
                "board_type": "B_1",
                "board_name": "Board 1",
                "product_code": "P1",
                "target_type": "platform",
                "slug": "Le Slug",
            }
        )

        self.assertEqual("B_1", board.board_type)
        self.assertEqual("Board 1", board.board_name)
        self.assertEqual(("Mbed OS 5.15",), board.mbed_os_support)
        self.assertEqual(("Basic",), board.mbed_enabled)
        self.assertEqual("P1", board.product_code)
        self.assertEqual("platform", board.target_type)
        self.assertEqual("Le Slug", board.slug)
        self.assertEqual((), board.build_variant)

    def test_empty_database_entry(self):
        """Given no data, a Board is created with no information."""
        board = Board.from_online_board_entry({})

        self.assertEqual("", board.board_type)
        self.assertEqual("", board.board_name)
        self.assertEqual((), board.mbed_os_support)
        self.assertEqual((), board.mbed_enabled)
        self.assertEqual("", board.product_code)
        self.assertEqual("", board.target_type)
        self.assertEqual("", board.slug)

    def test_online_database_entry(self):
        online_data = {
            "type": "target",
            "id": "1",
            "attributes": {
                "features": {
                    "mbed_enabled": ["Advanced"],
                    "mbed_os_support": [
                        "Mbed OS 5.10",
                        "Mbed OS 5.11",
                        "Mbed OS 5.12",
                        "Mbed OS 5.13",
                        "Mbed OS 5.14",
                        "Mbed OS 5.15",
                        "Mbed OS 5.8",
                        "Mbed OS 5.9",
                    ],
                    "antenna": ["Connector", "Onboard"],
                    "certification": [
                        "Anatel (Brazil)",
                        "AS/NZS (Australia and New Zealand)",
                        "CE (Europe)",
                        "FCC/CFR (USA)",
                        "IC RSS (Canada)",
                        "ICASA (South Africa)",
                        "KCC (South Korea)",
                        "MIC (Japan)",
                        "NCC (Taiwan)",
                        "RoHS (Europe)",
                    ],
                    "communication": ["Bluetooth & BLE"],
                    "interface_firmware": ["DAPLink", "J-Link"],
                    "target_core": ["Cortex-M4"],
                    "mbed_studio_support": ["Build and run"],
                },
                "board_type": "k64f",
                "flash_size": 512,
                "name": "u-blox NINA-B1",
                "product_code": "0455",
                "ram_size": 64,
                "target_type": "module",
                "hidden": False,
                "device_name": "nRF52832_xxAA",
                "slug": "u-blox-nina-b1",
            },
        }
        board = Board.from_online_board_entry(online_data)

        self.assertEqual(online_data["attributes"]["board_type"].upper(), board.board_type)
        self.assertEqual(online_data["attributes"]["name"], board.board_name)
        self.assertEqual(tuple(online_data["attributes"]["features"]["mbed_os_support"]), board.mbed_os_support)
        self.assertEqual(tuple(online_data["attributes"]["features"]["mbed_enabled"]), board.mbed_enabled)
        self.assertEqual(online_data["attributes"]["product_code"], board.product_code)
        self.assertEqual(online_data["attributes"]["target_type"], board.target_type)
        self.assertEqual(online_data["attributes"]["slug"], board.slug)
        self.assertEqual(tuple(), board.build_variant)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Command to build an Mbed project using CMake."""
import os
import pathlib
import shutil

import click

from mbed_tools.build import build_project, generate_build_system, generate_config
from mbed_tools.project import MbedProgram


@click.command(name="build", help="Build an Mbed project.")
@click.option(
    "-t",
    "--toolchain",
    type=click.Choice(["ARM", "GCC_ARM"], case_sensitive=False),
    help="The toolchain you are using to build your app.",
)
@click.option("-m", "--mbed-target", help="A build target for an Mbed-enabled device, e.g. K64F.")
@click.option("-b", "--build-type", default="develop", help="The build type (release, develop or debug).")
@click.option("-c", "--clean", is_flag=True, default=False, help="Perform a clean build.")
@click.option(
    "-p",
    "--program-path",
    default=os.getcwd(),
    help="Path to local Mbed program. By default it is the current working directory.",
)
def build(program_path: str, build_type: str, toolchain: str = "", mbed_target: str = "", clean: bool = False) -> None:
    """Configure and build an Mbed project using CMake and Ninja.

    If the project has already been configured and contains '.mbedbuild/mbed_config.cmake', this command will skip the
    Mbed configuration step and invoke CMake.

    If the CMake configuration step has already been run previously (i.e a CMake build tree exists), then just try to
    build the project immediately using Ninja.

    Args:
       program_path: Path to the Mbed project.
       build_type: The Mbed build profile (debug, develop or release).
       toolchain: The toolchain to use for the build.
       mbed_target: The name of the Mbed target to build for.
       clean: Perform a clean build.
    """
    program = MbedProgram.from_existing(pathlib.Path(program_path))
    mbed_config_file = program.files.cmake_config_file
    build_tree = program.files.cmake_build_dir
    if clean and build_tree.exists():
        shutil.rmtree(build_tree)

    if any([not mbed_config_file.exists(), not build_tree.exists(), mbed_target, toolchain]):
        click.echo("Configuring project and generating build system...")
        _validate_target_and_toolchain_args(mbed_target, toolchain)
        generate_config(mbed_target.upper(), toolchain, program)
        generate_build_system(program.root, build_tree, build_type)

    click.echo("Building Mbed project...")
    build_project(build_tree)


def _validate_target_and_toolchain_args(target: str, toolchain: str) -> None:
    if not all([toolchain, target]):
        raise click.UsageError("--toolchain and --mbed-target arguments are required when generating Mbed config!")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Package for parsing Mbed OS library's targets.json.

targets.json contains the attribute data for all supported targets.

Targets are defined using inheritance. Multiple inheritance is allowed (though not encouraged).

Attributes come in two types - overriding and accumulating. These
use two completely different ways of inheriting so each has its own parser.

The combined results of these two parsers makes a complete set of attributes for
a target as defined in targets.json.

For more information about targets.json structure see
https://os.mbed.com/docs/mbed-os/latest/reference/adding-and-configuring-targets.html
"""



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a Disk drive."""

from typing import NamedTuple, cast, Optional, Tuple
import re

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor
from mbed_tools.devices._internal.windows.component_descriptor_utils import is_undefined_value, UNKNOWN_VALUE
from mbed_tools.devices._internal.windows.windows_identifier import WindowsUID

PATTERN_UID = re.compile(r"[&#]?([0-9A-Za-z]{10,48})[&#]?")


class DiskDriveMsdnDefinition(NamedTuple):
    """Msdn definition of a disk drive.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/win32-diskdrive
    """

    Availability: int
    BytesPerSector: int
    Capabilities: list
    CapabilityDescriptions: list
    Caption: str
    CompressionMethod: str
    ConfigManagerErrorCode: int
    ConfigManagerUserConfig: bool
    CreationClassName: str
    DefaultBlockSize: int
    Description: str
    DeviceID: str
    ErrorCleared: bool
    ErrorDescription: str
    ErrorMethodology: str
    FirmwareRevision: str
    Index: int
    InstallDate: int
    InterfaceType: str
    LastErrorCode: int
    Manufacturer: str
    MaxBlockSize: int
    MaxMediaSize: int
    MediaLoaded: bool
    MediaType: str
    MinBlockSize: int
    Model: str
    Name: str
    NeedsCleaning: bool
    NumberOfMediaSupported: int
    Partitions: int
    PNPDeviceID: str
    PowerManagementCapabilities: list
    PowerManagementSupported: bool
    SCSIBus: int
    SCSILogicalUnit: int
    SCSIPort: int
    SCSITargetId: int
    SectorsPerTrack: int
    SerialNumber: str
    Signature: int
    Size: int
    Status: str
    StatusInfo: int
    SystemCreationClassName: str
    SystemName: str
    TotalCylinders: int
    TotalHeads: int
    TotalSectors: int
    TotalTracks: int
    TracksPerCylinder: int


class DiskDrive(ComponentDescriptor):
    """Disk Drive as defined in Windows API.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/win32-diskdrive
    """

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(DiskDriveMsdnDefinition, win32_class_name="Win32_DiskDrive")

    @property
    def component_id(self) -> str:
        """Returns the device id field."""
        return cast(str, self.get("DeviceID"))

    @property
    def uid(self) -> WindowsUID:
        """Returns the disk UID."""
        return Win32DiskIdParser().parse(cast(str, self.get("PNPDeviceID")), self.get("SerialNumber"))


class Win32DiskIdParser:
    """Parser of a standard Win32 Disk."""

    def _parse_pnpid(self, pnpid: str) -> Tuple[str, str]:
        if is_undefined_value(pnpid):
            return (UNKNOWN_VALUE, UNKNOWN_VALUE)
        parts = pnpid.split("\\")
        if len(parts) >= 2:
            raw_id = parts[-1]
            match = PATTERN_UID.search(raw_id)
            processed_uid = raw_id
            if match:
                processed_uid = match.group(1)
            return (processed_uid.lower(), raw_id)
        return (UNKNOWN_VALUE, UNKNOWN_VALUE)

    def parse(self, pnpid: str, serial_number: Optional[str]) -> WindowsUID:
        """Parses the UID value based on multiple fields.

        For different boards, the ID is stored in different fields.
        e.g. JLink serial number is irrelevant whereas it is the correct field for Daplink boards.
        For others, the PNPDeviceID should be used instead.
        """
        (uid, raw_uid) = self._parse_pnpid(pnpid)
        return WindowsUID(uid=uid, raw_uid=raw_uid, serial_number=serial_number)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import logging
from unittest import TestCase, mock

from mbed_tools.lib.exceptions import ToolsError
from mbed_tools.lib.logging import _exception_message, MbedToolsHandler, log_exception, set_log_level, LOGGING_FORMAT


class SubclassedToolsError(ToolsError):
    """An exception subclassing ToolsError."""


class TestExceptionMessage(TestCase):
    def test_critical_log_level_with_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.CRITICAL, True)
        self.assertTrue("'-v'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_critical_log_level_without_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.CRITICAL, False)
        self.assertTrue("'-v'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_error_log_level_with_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.ERROR, True)
        self.assertTrue("'-v'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_error_log_level_without_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.ERROR, False)
        self.assertTrue("'-v'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_warning_log_level_with_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.WARNING, True)
        self.assertTrue("'-vv'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_warning_log_level_without_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.WARNING, False)
        self.assertTrue("'-vv'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_info_log_level_with_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.INFO, True)
        self.assertTrue("'-vvv'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_info_log_level_without_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.INFO, False)
        self.assertTrue("'-vvv'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_debug_log_level_with_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.DEBUG, True)
        self.assertTrue("-v" not in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_debug_log_level_without_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.DEBUG, False)
        self.assertTrue("-v" not in message)
        self.assertTrue("'--traceback'" in message)
        self.assertTrue("unlikely string" in message)

    def test_log_level_not_set_with_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.NOTSET, True)
        self.assertTrue("'-v'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)

    def test_log_level_not_set_without_traceback(self):
        message = _exception_message(ToolsError("unlikely string"), logging.NOTSET, False)
        self.assertTrue("'-v'" in message)
        self.assertTrue("'--traceback'" not in message)
        self.assertTrue("unlikely string" in message)


class TestMbedToolsHandler(TestCase):
    exception_string: str = "A Message"
    expected_log_message: str = "A Message\n\nMore information may be available by using the command line option '-vv'."

    def test_no_exception_raised(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        with MbedToolsHandler(mock_logger, traceback=False):
            pass
        self.assertFalse(mock_logger.error.called, "Error should not be logger when an exception is not raised.")

    def test_tools_error_with_traceback(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        with MbedToolsHandler(mock_logger, traceback=True):
            raise ToolsError(self.exception_string)
        mock_logger.error.assert_called_once_with(self.expected_log_message, exc_info=True)

    def test_tools_error_without_traceback(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        with MbedToolsHandler(mock_logger, traceback=False):
            raise ToolsError(self.exception_string)
        mock_logger.error.assert_called_once_with(self.expected_log_message, exc_info=False)

    def test_subclassed_tools_error_with_traceback(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        with MbedToolsHandler(mock_logger, traceback=True):
            raise SubclassedToolsError(self.exception_string)
        mock_logger.error.assert_called_once_with(self.expected_log_message, exc_info=True)

    def test_subclassed_tools_error_without_traceback(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        with MbedToolsHandler(mock_logger, traceback=False):
            raise SubclassedToolsError(self.exception_string)
        mock_logger.error.assert_called_once_with(self.expected_log_message, exc_info=False)

    def test_other_exceptions(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        with self.assertRaises(ValueError):
            with MbedToolsHandler(mock_logger, traceback=False):
                raise ValueError(self.exception_string)
        self.assertFalse(mock_logger.error.called, "Error should not be logger when a tools error is not raised.")


class TestLogException(TestCase):
    def test_log_error(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        mock_exception = mock.Mock(spec_set=Exception)

        log_exception(mock_logger, mock_exception)

        mock_logger.error.assert_called_once_with(mock_exception, exc_info=False)

    def test_log_error_with_traceback(self):
        mock_logger = mock.Mock(spec_set=logging.Logger)
        mock_exception = mock.Mock(spec_set=Exception)

        log_exception(mock_logger, mock_exception, True)

        mock_logger.error.assert_called_once_with(mock_exception, exc_info=True)


@mock.patch("mbed_tools.lib.logging.logging", return_value=mock.Mock(spec_set=logging))
class TestSetLogLevel(TestCase):
    def test_debug(self, mocked_logging):
        set_log_level(verbose_count=3)
        mocked_logging.basicConfig.assert_called_once_with(level=mocked_logging.DEBUG, format=LOGGING_FORMAT)

    def test_info(self, mocked_logging):
        set_log_level(verbose_count=2)
        mocked_logging.basicConfig.assert_called_once_with(level=mocked_logging.INFO, format=LOGGING_FORMAT)

    def test_warning(self, mocked_logging):
        set_log_level(verbose_count=1)
        mocked_logging.basicConfig.assert_called_once_with(level=mocked_logging.WARNING, format=LOGGING_FORMAT)

    def test_error(self, mocked_logging):
        set_log_level(verbose_count=0)
        mocked_logging.basicConfig.assert_called_once_with(level=mocked_logging.ERROR, format=LOGGING_FORMAT)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Interactions with `system_profiler`."""
import plistlib
import re
import subprocess
from typing import Dict, Iterable, List, cast
from typing_extensions import TypedDict

USBDeviceTree = Dict  # mypy does not work with recursive types, which "_items" would require


class USBDeviceMedia(TypedDict, total=False):
    """Representation of usb device storage."""

    bsd_name: str


class USBDevice(TypedDict, total=False):
    """Representation of usb device."""

    _name: str
    location_id: str
    vendor_id: str
    product_id: str
    serial_num: str
    Media: List[USBDeviceMedia]


def get_all_usb_devices_data() -> List[USBDeviceTree]:
    """Returns parsed output of `system_profiler` call."""
    output = subprocess.check_output(["system_profiler", "-xml", "SPUSBDataType"], stderr=subprocess.DEVNULL)
    if output:
        return cast(List[USBDeviceTree], plistlib.loads(output))
    return []


def get_end_usb_devices_data() -> List[USBDevice]:
    """Returns only end devices from the output of `system_profiler` call."""
    data = get_all_usb_devices_data()
    leaf_devices = _extract_leaf_devices(data)
    end_devices = _filter_end_devices(leaf_devices)
    return end_devices


def _extract_leaf_devices(data: Iterable[USBDeviceTree]) -> List[USBDevice]:
    """Flattens the structure returned by `system_profiler` call.

    Expected input will contain a tree-like structures, this function will return their leaf nodes.
    """
    end_devices = []
    for device in data:
        if "_items" in device:
            child_devices = _extract_leaf_devices(device["_items"])
            end_devices.extend(child_devices)
        else:
            end_devices.append(cast(USBDevice, device))
    return end_devices


def _filter_end_devices(data: Iterable[USBDevice]) -> List[USBDevice]:
    """Removes devices that don't look like end devices.

    An end device is a device that shouldn't have child devices.
    I.e.: a hub IS NOT an end device, a mouse IS an end device.
    """
    return [device for device in data if not _is_hub(device) and not _is_bus(device)]


def _is_hub(data: USBDevice) -> bool:
    return bool(re.match(r"USB\d.\d Hub", data.get("_name", "")))


def _is_bus(data: USBDevice) -> bool:
    return bool(re.match(r"USB\d\dBus", data.get("_name", "")))



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib
from unittest import TestCase
from mbed_tools.devices._internal.candidate_device import CandidateDevice


def build_candidate_data(**overrides):
    defaults = {
        "product_id": "0x1234",
        "vendor_id": "0x5678",
        "mount_points": (pathlib.Path("./foo"),),
        "serial_number": "qwer",
        "serial_port": "COM1",
    }
    return {**defaults, **overrides}


class TestCandidateDevice(TestCase):
    def test_produces_a_valid_candidate(self):
        candidate_data = build_candidate_data()
        candidate = CandidateDevice(**candidate_data)

        self.assertEqual(candidate.product_id, candidate_data["product_id"])
        self.assertEqual(candidate.vendor_id, candidate_data["vendor_id"])
        self.assertEqual(candidate.mount_points, candidate_data["mount_points"])
        self.assertEqual(candidate.serial_number, candidate_data["serial_number"])
        self.assertEqual(candidate.serial_port, candidate_data["serial_port"])

    def test_raises_when_product_id_is_empty(self):
        candidate_data = build_candidate_data(product_id="")
        with self.assertRaisesRegex(ValueError, "product_id"):
            CandidateDevice(**candidate_data)

    def test_raises_when_product_id_is_not_hex(self):
        candidate_data = build_candidate_data(product_id="TEST")
        with self.assertRaisesRegex(ValueError, "product_id"):
            CandidateDevice(**candidate_data)

    def test_prefixes_product_id_hex_value(self):
        candidate_data = build_candidate_data(product_id="ff01")
        candidate = CandidateDevice(**candidate_data)
        self.assertEqual(candidate.product_id, "0xff01")

    def test_raises_when_vendor_id_is_empty(self):
        candidate_data = build_candidate_data(vendor_id="")
        with self.assertRaisesRegex(ValueError, "vendor_id"):
            CandidateDevice(**candidate_data)

    def test_raises_when_vendor_id_is_not_hex(self):
        candidate_data = build_candidate_data(vendor_id="TEST")
        with self.assertRaisesRegex(ValueError, "vendor_id"):
            CandidateDevice(**candidate_data)

    def test_prefixes_vendor_id_hex_value(self):
        candidate_data = build_candidate_data(vendor_id="cbad")
        candidate = CandidateDevice(**candidate_data)
        self.assertEqual(candidate.vendor_id, "0xcbad")

    def test_raises_when_mount_points_are_empty(self):
        with self.assertRaisesRegex(ValueError, "mount_points"):
            CandidateDevice(product_id="1234", vendor_id="1234", mount_points=[], serial_number="1234")

    def test_raises_when_serial_number_is_empty(self):
        candidate_data = build_candidate_data(serial_number="")
        with self.assertRaisesRegex(ValueError, "serial_number"):
            CandidateDevice(**candidate_data)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Device detector for Darwin."""
import logging
import pathlib
import re
from typing import List, Tuple, Optional
from typing_extensions import TypedDict
from mbed_tools.devices._internal.base_detector import DeviceDetector
from mbed_tools.devices._internal.candidate_device import CandidateDevice
from mbed_tools.devices._internal.darwin import system_profiler, ioreg, diskutil


logger = logging.getLogger(__name__)


class CandidateDeviceData(TypedDict):
    """CandidateDeviceData calculated from USBDevice."""

    vendor_id: str
    product_id: str
    serial_number: str
    mount_points: Tuple[pathlib.Path, ...]
    serial_port: Optional[str]


class InvalidCandidateDeviceDataError(ValueError):
    """Raised when CandidateDevice was given invalid data and it cannot be built."""

    pass


class DarwinDeviceDetector(DeviceDetector):
    """Darwin specific implementation of device detection."""

    def find_candidates(self) -> List[CandidateDevice]:
        """Return a list of CandidateDevices."""
        usb_devices_data = system_profiler.get_end_usb_devices_data()
        candidates = []
        for device_data in usb_devices_data:
            logging.debug(f"Building from: {device_data}.")
            try:
                candidate = _build_candidate(device_data)
            except InvalidCandidateDeviceDataError:
                pass
            else:
                logging.debug(f"Built candidate: {candidate}.")
                candidates.append(candidate)
        return candidates


def _build_candidate(device_data: system_profiler.USBDevice) -> CandidateDevice:
    assembled_data = _assemble_candidate_data(device_data)
    try:
        return CandidateDevice(**assembled_data)
    except ValueError as e:
        logging.debug(f"Unable to build candidate. {e}")
        raise InvalidCandidateDeviceDataError


def _assemble_candidate_data(device_data: system_profiler.USBDevice) -> CandidateDeviceData:
    return {
        "vendor_id": _format_vendor_id(device_data.get("vendor_id", "")),
        "product_id": device_data.get("product_id", ""),
        "serial_number": device_data.get("serial_num", ""),
        "mount_points": _get_mount_points(device_data),
        "serial_port": _get_serial_port(device_data),
    }


def _format_vendor_id(vendor_id: str) -> str:
    """Strips vendor name from vendor_id field.

    Example:
        >>> _format_vendor_id("0x1234 (Nice Vendor Inc.)")  # "0x1234"
    """
    return vendor_id.split(maxsplit=1)[0]


def _get_mount_points(device_data: system_profiler.USBDevice) -> Tuple[pathlib.Path, ...]:
    """Returns mount points for a given device, empty list if device has no mount points."""
    storage_identifiers = [media["bsd_name"] for media in device_data.get("Media", []) if "bsd_name" in media]
    mount_points = []
    for storage_identifier in storage_identifiers:
        mount_point = diskutil.get_mount_point(storage_identifier)
        if mount_point:
            mount_points.append(pathlib.Path(mount_point))
        else:
            logging.debug(f"Couldn't determine mount point for device id: {storage_identifier}.")
    return tuple(mount_points)


def _get_serial_port(device_data: system_profiler.USBDevice) -> Optional[str]:
    """Returns serial port for a given device, None if serial port cannot be determined."""
    device_name = device_data.get("_name")
    if not device_name:
        logging.debug('Missing "_name" in "{device_data}", which is required for ioreg name.')
        return None

    location_id = device_data.get("location_id")
    if not location_id:
        logging.debug('Missing "location_id" in "{device_data}", which is required for ioreg name.')
        return None

    ioreg_name = _build_ioreg_device_name(device_name=device_name, location_id=location_id)
    serial_port = ioreg.get_io_dialin_device(ioreg_name)
    return serial_port


def _build_ioreg_device_name(device_name: str, location_id: str) -> str:
    """Converts extracted `_name` and `location_id` attributes from `system_profiler` to a valid ioreg device name.

    `system_profiler` utility returns location ids in the form of `0xNNNNNNN`, with an optional suffix of ` / N`.

    Example:
        >>> _build_ioreg_device_name("STM32 Foo", "0x123456 / 2")
        "STM32 Foo@123456"
    """
    pattern = r"""
    0x                # hexadecimal prefix
    (?P<location>\d+) # location (i.e.: "123456" in "0x123456 / 2")
    (\s\/\s\d+)?      # suffix of location (" / 14")
    """
    match = re.match(pattern, location_id, re.VERBOSE)
    if match:
        return f"{device_name}@{match['location']}"
    else:
        return device_name



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Wrappers for git operations."""
from dataclasses import dataclass
from pathlib import Path

import git

from mbed_tools.project.exceptions import VersionControlError
from mbed_tools.project._internal.progress import ProgressReporter


@dataclass
class GitReference:
    """Git reference for a remote repository.

    Attributes:
        repo_url: URL of the git repository.
        ref: The reference commit sha, tag or branch.
    """

    repo_url: str
    ref: str


def clone(url: str, dst_dir: Path) -> git.Repo:
    """Clone a library repository.

    Args:
        url: URL of the remote to clone.
        dst_dir: Destination directory for the cloned repo.

    Raises:
        VersionControlError: Cloning the repository failed.
    """
    try:
        return git.Repo.clone_from(url, str(dst_dir), progress=ProgressReporter(name=url))
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Cloning git repository from url '{url}' failed. Error from VCS: {err.stderr}")


def checkout(repo: git.Repo, ref: str, force: bool = False) -> None:
    """Check out a specific reference in the given repository.

    Args:
        repo: git.Repo object where the checkout will be performed.
        ref: Git commit hash, branch or tag reference, must be a valid ref defined in the repo.

    Raises:
        VersionControlError: Check out failed.
    """
    try:
        repo.git.checkout(f"--force {ref}" if force else str(ref))
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Failed to check out revision '{ref}'. Error from VCS: {err.stderr}")


def init(path: Path) -> git.Repo:
    """Initialise a git repository at the given path.

    Args:
        path: Path where the repo will be initialised.

    Returns:
        Initialised git.Repo object.

    Raises:
        VersionControlError: initalising the repository failed.
    """
    try:
        return git.Repo.init(str(path))
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Failed to initialise git repository at path '{path}'. Error from VCS: {err.stderr}")


def get_repo(path: Path) -> git.Repo:
    """Get a git.Repo object from an existing repository path.

    Args:
        path: Path to the git repository.

    Returns:
        git.Repo object.

    Raises:
        VersionControlError: No valid git repository at this path.
    """
    try:
        return git.Repo(str(path))
    except git.exc.InvalidGitRepositoryError:
        raise VersionControlError(
            "Could not find a valid git repository at this path. Please perform a `git init` command."
        )



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Public exceptions raised by the package."""
from mbed_tools.lib.exceptions import ToolsError


class MbedDevicesError(ToolsError):
    """Base public exception for the mbed-devices package."""


class DeviceLookupFailed(MbedDevicesError):
    """Failed to look up data associated with the device."""



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Utilities with regards to Win32 component descriptors."""
from typing import Any, NamedTuple, Union
from collections import OrderedDict

UNKNOWN_VALUE = "Unknown"


def is_undefined_value(value: Any) -> bool:
    """Checks whether a field value is considered as undefined or not."""
    return value in [None, UNKNOWN_VALUE, 0, False]


def is_undefined_data_object(data_object: NamedTuple) -> bool:
    """Checks whether a data object is considered as undefined or not."""
    for f in data_object._fields:
        if not is_undefined_value(getattr(data_object, f)):
            is_undefined = False
            break
    else:
        is_undefined = True
    return is_undefined


def retain_value_or_default(value: Any) -> Any:
    """Returns the value if set otherwise returns a default value."""
    return UNKNOWN_VALUE if is_undefined_value(value) else value


def data_object_to_dict(data_object: NamedTuple) -> dict:
    """Gets a data object as a dictionary."""
    internal_dictionary: Union[dict, OrderedDict] = data_object._asdict()
    return dict(internal_dictionary) if isinstance(internal_dictionary, OrderedDict) else internal_dictionary



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from pathlib import Path
from unittest import TestCase, mock

from mbed_tools.project.exceptions import VersionControlError
from mbed_tools.project._internal import git_utils


class TestClone(TestCase):
    @mock.patch("mbed_tools.project._internal.git_utils.git.Repo", autospec=True)
    @mock.patch("mbed_tools.project._internal.git_utils.ProgressReporter", autospec=True)
    def test_returns_repo(self, mock_progress, mock_repo):
        url = "https://blah"
        path = Path()
        repo = git_utils.clone(url, path)

        self.assertIsNotNone(repo)
        mock_repo.clone_from.assert_called_once_with(url, str(path), progress=mock_progress())

    def test_raises_when_clone_fails(self):
        with self.assertRaises(VersionControlError):
            git_utils.clone("", Path())


@mock.patch("mbed_tools.project._internal.git_utils.git.Repo", autospec=True)
class TestInit(TestCase):
    def test_returns_initialised_repo(self, mock_repo):
        repo = git_utils.init(Path())

        self.assertIsNotNone(repo)
        mock_repo.init.assert_called_once_with(str(Path()))

    def test_raises_when_init_fails(self, mock_repo):
        mock_repo.init.side_effect = git_utils.git.exc.GitCommandError("git init", 255)

        with self.assertRaises(VersionControlError):
            git_utils.init(Path())


@mock.patch("mbed_tools.project._internal.git_utils.git.Repo", autospec=True)
class TestGetRepo(TestCase):
    def test_returns_repo_object(self, mock_repo):
        repo = git_utils.get_repo(Path())

        self.assertTrue(isinstance(repo, mock_repo.__class__))

    def test_raises_version_control_error_when_no_git_repo_found(self, mock_repo):
        mock_repo.side_effect = git_utils.git.exc.InvalidGitRepositoryError

        with self.assertRaises(VersionControlError):
            git_utils.get_repo(Path())


@mock.patch("mbed_tools.project._internal.git_utils.git.Repo", autospec=True)
class TestCheckout(TestCase):
    def test_git_lib_called_with_correct_command(self, mock_repo):
        git_utils.checkout(mock_repo, "master", force=True)

        mock_repo.git.checkout.assert_called_once_with("--force master")

    def test_raises_version_control_error_when_git_checkout_fails(self, mock_repo):
        mock_repo.git.checkout.side_effect = git_utils.git.exc.GitCommandError("git checkout", 255)

        with self.assertRaises(VersionControlError):
            git_utils.checkout(mock_repo, "bad")



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines the public API of the package."""
import pathlib
import logging

from typing import Dict, Any

from mbed_tools.project.mbed_program import MbedProgram, parse_url

logger = logging.getLogger(__name__)


def clone_project(url: str, dst_path: Any = None, recursive: bool = False) -> None:
    """Clones an Mbed project from a remote repository.

    Args:
        url: URL of the repository to clone.
        dst_path: Destination path for the repository.
        recursive: Recursively clone all project dependencies.
    """
    git_data = parse_url(url)
    url = git_data["url"]
    if not dst_path:
        dst_path = pathlib.Path(git_data["dst_path"])

    program = MbedProgram.from_url(url, dst_path)
    if recursive:
        program.resolve_libraries()


def initialise_project(path: pathlib.Path, create_only: bool) -> None:
    """Create a new Mbed project, optionally fetching and adding mbed-os.

    Args:
        path: Path to the project folder. Created if it doesn't exist.
        create_only: Flag which suppreses fetching mbed-os. If the value is `False`, fetch mbed-os from the remote.
    """
    program = MbedProgram.from_new(path)
    if not create_only:
        program.resolve_libraries()


def deploy_project(path: pathlib.Path, force: bool = False) -> None:
    """Deploy a specific revision of the current Mbed project.

    This function also resolves and syncs all library dependencies to the revision specified in the library reference
    files.

    Args:
        path: Path to the Mbed project.
        project_revision: Revision of the Mbed project to check out.
        force: Force overwrite uncommitted changes. If False, the deploy will fail if there are uncommitted local
               changes.
    """
    program = MbedProgram.from_existing(path, check_mbed_os=False)
    program.deploy_libraries(force=force)
    if program.has_unresolved_libraries():
        logger.info("Unresolved libraries detected, downloading library source code.")
        program.resolve_libraries()


def get_known_libs(path: pathlib.Path) -> Dict[str, Any]:
    """List all resolved library dependencies.

    This function will not resolve dependencies. This will only generate a list of resolved dependencies.

    Args:
        path: Path to the Mbed project.

    Returns:
        dictionary containing a list of known dependencies and a boolean stating whether unresolved dependencies were
        detected.
    """
    program = MbedProgram.from_existing(path, check_mbed_os=False)

    return {"known_libs": program.list_known_library_dependencies(), "unresolved": program.has_unresolved_libraries()}



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Module in charge of CMake file generation."""
import pathlib
from typing import Iterable

import jinja2

from mbed_tools.build._internal.config.config import Config

TEMPLATES_DIRECTORY = pathlib.Path("_internal", "templates")
TEMPLATE_NAME = "mbed_config.tmpl"


def generate_mbed_config_cmake_file(
    mbed_target_name: str, target_build_attributes: dict, config: Config, toolchain_name: str
) -> str:
    """Generate the top-level CMakeLists.txt file containing the correct definitions for a build.

    Args:
        mbed_target: the target the application is being built for
        target_build_attributes: map of target attributes
        program_path: the path to the local Mbed program
        toolchain_name: the toolchain to be used to build the application

    Returns:
        A string of rendered contents for the file.
    """
    return _render_mbed_config_cmake_template(target_build_attributes, config, toolchain_name, mbed_target_name,)


def _render_mbed_config_cmake_template(
    target_build_attributes: dict, config: Config, toolchain_name: str, target_name: str
) -> str:
    """Renders the mbed_config template with the relevant information.

    Args:
        target_build_attributes: Target config object.
        config: Config object holding information parsed from the mbed config system.
        toolchain_name: Name of the toolchain being used.
        target_name: Name of the target.

    Returns:
        The contents of the rendered CMake file.
    """
    env = jinja2.Environment(loader=jinja2.PackageLoader("mbed_tools.build", str(TEMPLATES_DIRECTORY)),)
    template = env.get_template(TEMPLATE_NAME)
    options = list(config.options.values())
    macros = list(config.macros.values())
    supported_c_libs = [x for x in target_build_attributes["supported_c_libs"][toolchain_name.lower()]]

    context = {
        "labels": target_build_attributes["labels"],
        "extra_labels": target_build_attributes["extra_labels"],
        "features": target_build_attributes["features"],
        "components": target_build_attributes["components"],
        "device_has": target_build_attributes["device_has"],
        "target_macros": target_build_attributes["macros"],
        "supported_form_factors": target_build_attributes["supported_form_factors"],
        "supported_c_libs": supported_c_libs,
        "supported_application_profiles": target_build_attributes["supported_application_profiles"],
        "c_lib": target_build_attributes["c_lib"],
        "core": target_build_attributes["core"],
        "printf_lib": target_build_attributes["printf_lib"],
        "target_name": target_name,
        "toolchain_name": toolchain_name,
        "options": sorted(options, key=lambda option: option.macro_name),
        "macros": sorted(macros, key=lambda macro: macro.name),
        "max_name_length": max(_max_attribute_length(options, "macro_name"), _max_attribute_length(macros, "name")),
        "max_value_length": max(_max_attribute_length(options, "value"), _max_attribute_length(macros, "value")),
    }
    return template.render(context)


def _max_attribute_length(objects: Iterable[object], attribute: str) -> int:
    attrs = (getattr(o, attribute) for o in objects)
    try:
        return max(len(str(attr)) for attr in attrs if attr is not None)
    except ValueError:  # no attrs found
        return 0



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Configuration source abstraction."""
import json
import logging

from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Any

logger = logging.getLogger(__name__)


@dataclass
class Source:
    """Configuration source abstraction.

    MbedOS build configuration is assembled from various sources:
    - targets.json
    - mbed_lib.json
    - mbed_app.json

    This class provides a common interface to configuration data.
    """

    human_name: str
    config: dict
    overrides: dict
    macros: Iterable[str]

    @classmethod
    def from_mbed_lib(cls, mbed_lib_path: Path, target_labels: Iterable[str]) -> "Source":
        """Build Source from mbed_lib.json file.

        Args:
            mbed_lib_path: Path to mbed_lib.json file
            target_labels: Labels for which "target_overrides" should apply
        """
        file_contents = _decode_json_file(mbed_lib_path)
        namespace = file_contents["name"]

        return cls.from_file_contents(
            file_name=str(mbed_lib_path), file_contents=file_contents, namespace=namespace, target_labels=target_labels
        )

    @classmethod
    def from_mbed_app(cls, mbed_app_path: Path, target_labels: Iterable[str]) -> "Source":
        """Build Source from mbed_app.json file.

        Args:
            mbed_app_path: Path to mbed_app.json file
            target_labels: Labels for which "target_overrides" should apply
        """
        file_contents = _decode_json_file(mbed_app_path)
        return cls.from_file_contents(
            file_name=str(mbed_app_path), file_contents=file_contents, namespace="app", target_labels=target_labels
        )

    @classmethod
    def from_file_contents(
        cls, file_name: str, file_contents: dict, namespace: str, target_labels: Iterable[str]
    ) -> "Source":
        """Build Source from file contents."""
        config = file_contents.get("config", {})
        config = _namespace_data(config, namespace)

        overrides = file_contents.get("target_overrides", {})
        target_specific_overrides = _filter_target_overrides(overrides, target_labels)
        namespaced_overrides = _namespace_data(target_specific_overrides, namespace)

        macros = file_contents.get("macros", [])

        return cls(human_name=f"File: {file_name}", config=config, overrides=namespaced_overrides, macros=macros)

    @classmethod
    def from_target(cls, target: dict) -> "Source":
        """Build Source from retrieved mbed_tools.targets.Target data."""
        namespace = "target"
        config = _namespace_data(target["config"], namespace)

        overrides = {
            "features": target["features"],
            "components": target["components"],
            "labels": target["labels"],
            "extra_labels": target["extra_labels"],
            "macros": target["macros"],
        }
        namespaced_overrides = _namespace_data(overrides, namespace)

        return cls(
            human_name=f"mbed_target.Target for {target}", config=config, overrides=namespaced_overrides, macros=[],
        )


def _filter_target_overrides(data: dict, allowed_labels: Iterable[str]) -> dict:
    """Flatten and filter target overrides.

    Ensures returned dictionary only contains configuration settings applicable to given allowed labels.
    """
    flattened = {}
    for target_label, overrides in data.items():
        if target_label == "*" or target_label in allowed_labels:
            flattened.update(overrides)
    return flattened


def _namespace_data(data: dict, namespace: str) -> dict:
    """Prefix configuration key with a namespace.

    Namespace is ConfigSource wide, and is resolved at source build time.

    It should be one of:
    - "target"
    - "app"
    - library name (where "mbed_lib.json" comes from)

    If given key is already namespaced, return it as is - this is going to be the case for
    keys from "target_overrides" entries. Keys from "config" usually need namespacing.
    """
    namespaced = {}
    for key, value in data.items():
        if "." not in key:
            key = f"{namespace}.{key}"
        namespaced[key] = value
    return namespaced


def _decode_json_file(path: Path) -> Any:
    try:
        return json.loads(path.read_text())
    except json.JSONDecodeError:
        logger.error(f"Failed to decode JSON data in the file located at '{path}'")
        raise



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase
from mbed_tools.devices._internal.windows.component_descriptor_utils import data_object_to_dict
from mbed_tools.devices._internal.windows.windows_identifier import WindowsUID
from tests.devices._internal.windows.test_windows_identifier import generateUID
from tests.devices.markers import windows_only


@windows_only
class TestUsbDeviceId(TestCase):
    """Tests based on https://docs.microsoft.com/en-us/windows-hardware/drivers/install/standard-usb-identifiers."""

    def test_single_interface_usb_device(self):
        from mbed_tools.devices._internal.windows.usb_device_identifier import parse_device_id

        self.assertTrue(parse_device_id("").is_undefined)
        self.assertTrue(parse_device_id(None).is_undefined)
        self.assertTrue(parse_device_id("4&38EF038C&0&0").is_undefined)
        self.assertFalse(parse_device_id("USB\\4&38EF038C&0&0").is_undefined)
        self.assertEqual(
            parse_device_id("USB\\4&38EF038C&0&0").uid,
            WindowsUID(uid="4&38ef038c&0", raw_uid="4&38EF038C&0&0", serial_number=None),
        )
        self.assertEqual(parse_device_id("USB\\ROOT_HUB30\\4&38EF038C&0&0").uid.raw_uid, "4&38EF038C&0&0")
        self.assertEqual(
            parse_device_id("USB\\VID_2109&PID_2812\\6&38E4CCB6&0&4").uid,
            WindowsUID(uid="6&38e4ccb6&0", raw_uid="6&38E4CCB6&0&4", serial_number=None),
        )
        self.assertEqual(parse_device_id("USB\\VID_2109&PID_2812\\6&38E4CCB6&0&4").PID, "2812")
        self.assertEqual(parse_device_id("USB\\VID_2109&PID_2812\\6&38E4CCB6&0&4").VID, "2109")
        self.assertEqual(parse_device_id("USB\\VID_2109&PID_2812&REV_1100\\6&38E4CCB6&0&4").REV, "1100")

        self.assertGreaterEqual(
            data_object_to_dict(parse_device_id("USB\\VID_2109&PID_2812&REV_1100\\6&38E4CCB6&0&4")).items(),
            {
                "VID": "2109",
                "PID": "2812",
                "REV": "1100",
                "UID": WindowsUID(uid="6&38e4ccb6&0", raw_uid="6&38E4CCB6&0&4", serial_number=None),
            }.items(),
        )

    def test_multiple_interface_usb_device(self):
        from mbed_tools.devices._internal.windows.usb_device_identifier import parse_device_id

        self.assertEqual(
            parse_device_id("USB\\VID_0D28&PID_0204&MI_00\\0240000034544E45001A00018AA900292011000097969900").uid.uid,
            "0240000034544e45001a00018aa900292011000097969900",
        )
        self.assertEqual(
            parse_device_id(
                "USB\\VID_0D28&PID_0204&MI_00\\0240000034544E45001A00018AA900292011000097969900"
            ).uid.raw_uid,
            "0240000034544E45001A00018AA900292011000097969900",
        )
        self.assertEqual(
            parse_device_id("USB\\VID_0D28&PID_0204&MI_00\\0240000034544E45001A00018AA900292011000097969900").PID,
            "0204",
        )
        self.assertEqual(
            parse_device_id("USB\\VID_0D28&PID_0204&MI_02\\0240000034544E45001A00018AA900292011000097969900").VID,
            "0D28",
        )
        self.assertEqual(
            parse_device_id("USB\\VID_0D28&PID_0204&MI_02\\0240000034544E45001A00018AA900292011000097969900").MI, "02"
        )

    def test_equals(self):
        from mbed_tools.devices._internal.windows.usb_device_identifier import UsbIdentifier, KEY_UID

        # Checks that two unset identifiers are equal
        a = UsbIdentifier()
        b = UsbIdentifier()
        self.assertEqual(a, a)
        self.assertEqual(a, b)
        self.assertEqual(b, a)
        # Checks that two different identifiers are not equal
        a_dict = data_object_to_dict(a)
        a_dict[KEY_UID] = generateUID()
        b_dict = data_object_to_dict(a)
        b_dict[KEY_UID] = generateUID()
        a = UsbIdentifier(**a_dict)
        b = UsbIdentifier(**b_dict)
        self.assertEqual(a, a)
        self.assertNotEqual(a.uid, b.uid)
        self.assertNotEqual(a, b)
        self.assertNotEqual(b, a)
        # Checks that two identifiers with same fields are equal
        b = UsbIdentifier(**a_dict)
        self.assertEqual(a.uid, b.uid)
        self.assertEqual(a, a)
        self.assertEqual(a, b)
        self.assertEqual(b, a)

    def test_hashing(self):
        from mbed_tools.devices._internal.windows.usb_device_identifier import UsbIdentifier, KEY_UID

        # Generates two different USB identifiers
        a = UsbIdentifier()
        a_dict = data_object_to_dict(a)
        a_dict[KEY_UID] = generateUID()
        b_dict = data_object_to_dict(a)
        b_dict[KEY_UID] = generateUID()
        a = UsbIdentifier(**a_dict)
        b = UsbIdentifier(**b_dict)

        self.assertNotEqual(hash(a), hash(b))
        self.assertNotEqual(a, b)
        # Checks dictionary lookup
        self.assertNotIn(a, dict())
        self.assertNotIn(a, {b: ""})
        self.assertIn(b, {b: ""})

        # Creates c so that a and c have the same fields.
        c = UsbIdentifier(**a_dict)

        # Checks dictionary lookup
        self.assertIn(c, {c: ""})
        self.assertIn(a, {c: ""})



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for parsing the attributes for targets in targets.json that accumulate."""
from unittest import TestCase, mock

from mbed_tools.targets._internal.targets_json_parsers.accumulating_attribute_parser import (
    ALL_ACCUMULATING_ATTRIBUTES,
    get_accumulating_attributes_for_target,
    _targets_accumulate_hierarchy,
    _determine_accumulated_attributes,
    _remove_attribute_element,
    _element_matches,
)


class ListAllAccumulatingAttributes(TestCase):
    def test_expected_list(self):
        expected_attributes = (
            "extra_labels",
            "macros",
            "device_has",
            "features",
            "components",
            "extra_labels_add",
            "extra_labels_remove",
            "macros_add",
            "macros_remove",
            "device_has_add",
            "device_has_remove",
            "features_add",
            "features_remove",
            "components_add",
            "components_remove",
        )
        self.assertEqual(ALL_ACCUMULATING_ATTRIBUTES, expected_attributes)


class TestGetAccumulatingAttributes(TestCase):
    @mock.patch(
        "mbed_tools.targets._internal.targets_json_parsers."
        "accumulating_attribute_parser._targets_accumulate_hierarchy"
    )
    @mock.patch(
        "mbed_tools.targets._internal.targets_json_parsers."
        "accumulating_attribute_parser._determine_accumulated_attributes"
    )
    def test_correctly_calls(self, _determine_accumulated_attributes, _targets_accumulate_hierarchy):
        target_name = "Target_Name"
        all_targets_data = {target_name: {"attribute_1": ["something"]}}
        result = get_accumulating_attributes_for_target(all_targets_data, target_name)

        _targets_accumulate_hierarchy.assert_called_once_with(all_targets_data, target_name)
        _determine_accumulated_attributes.assert_called_once_with(_targets_accumulate_hierarchy.return_value)
        self.assertEqual(result, _determine_accumulated_attributes.return_value)


class TestParseHierarchy(TestCase):
    def test_accumulate_hierarchy_single_inheritance(self):
        all_targets_data = {
            "D": {"attribute_1": ["some things"]},
            "C": {"inherits": ["D"], "attribute_2": "something else"},
            "B": {},
            "A": {"inherits": ["C"], "attribute_3": ["even more things"]},
        }
        result = _targets_accumulate_hierarchy(all_targets_data, "A")

        self.assertEqual(result, [all_targets_data["A"], all_targets_data["C"], all_targets_data["D"]])

    def test_accumulate_hierarchy_multiple_inheritance(self):
        all_targets_data = {
            "F": {"attribute_1": "some thing"},
            "E": {"attribute_2": "some other thing"},
            "D": {"inherits": ["F"]},
            "C": {"inherits": ["E"]},
            "B": {"inherits": ["C", "D"]},
            "A": {"inherits": ["B"]},
        }
        result = _targets_accumulate_hierarchy(all_targets_data, "A")

        self.assertEqual(
            result,
            [
                all_targets_data["A"],
                all_targets_data["B"],
                all_targets_data["C"],
                all_targets_data["D"],
                all_targets_data["E"],
                all_targets_data["F"],
            ],
        )


class TestAccumulatingAttributes(TestCase):
    def test_determine_accumulated_attributes_basic_add(self):
        accumulation_order = [
            {"attribute_1": "something"},
            {f"{ALL_ACCUMULATING_ATTRIBUTES[0]}_add": ["2", "3"]},
            {ALL_ACCUMULATING_ATTRIBUTES[0]: ["1"]},
        ]
        expected_attributes = {ALL_ACCUMULATING_ATTRIBUTES[0]: ["1", "2", "3"]}
        result = _determine_accumulated_attributes(accumulation_order)
        self.assertEqual(result, expected_attributes)

    def test_determine_accumulated_attributes_basic_remove(self):
        accumulation_order = [
            {"attribute_1": "something"},
            {f"{ALL_ACCUMULATING_ATTRIBUTES[0]}_remove": ["2", "3"]},
            {ALL_ACCUMULATING_ATTRIBUTES[0]: ["1", "2", "3"]},
        ]
        expected_attributes = {ALL_ACCUMULATING_ATTRIBUTES[0]: ["1"]}
        result = _determine_accumulated_attributes(accumulation_order)
        self.assertEqual(result, expected_attributes)

    def test_combination_multiple_attributes(self):
        accumulation_order = [
            {f"{ALL_ACCUMULATING_ATTRIBUTES[0]}_add": ["2", "3"]},
            {f"{ALL_ACCUMULATING_ATTRIBUTES[1]}_remove": ["B", "C"]},
            {ALL_ACCUMULATING_ATTRIBUTES[0]: ["1"]},
            {ALL_ACCUMULATING_ATTRIBUTES[1]: ["A", "B", "C"]},
        ]
        expected_attributes = {
            ALL_ACCUMULATING_ATTRIBUTES[0]: ["1", "2", "3"],
            ALL_ACCUMULATING_ATTRIBUTES[1]: ["A"],
        }
        result = _determine_accumulated_attributes(accumulation_order)
        self.assertEqual(result, expected_attributes)

    def test_combination_later_check_no_unwanted_overrides(self):
        accumulation_order = [
            {f"{ALL_ACCUMULATING_ATTRIBUTES[0]}_add": ["2", "3"]},
            {f"{ALL_ACCUMULATING_ATTRIBUTES[1]}_remove": ["B", "C"]},
            {ALL_ACCUMULATING_ATTRIBUTES[0]: ["1"]},
            {ALL_ACCUMULATING_ATTRIBUTES[1]: ["A", "B", "C"]},
            {ALL_ACCUMULATING_ATTRIBUTES[1]: []},
        ]
        expected_attributes = {
            ALL_ACCUMULATING_ATTRIBUTES[0]: ["1", "2", "3"],
            ALL_ACCUMULATING_ATTRIBUTES[1]: ["A"],
        }
        result = _determine_accumulated_attributes(accumulation_order)
        self.assertEqual(result, expected_attributes)


class TestElementMatches(TestCase):
    def test_element_matches_exactly(self):
        element_to_remove = "SOMETHING"
        element_to_check = "SOMETHING"

        self.assertTrue(_element_matches(element_to_remove, element_to_check))

    def test_element_no_match(self):
        element_to_remove = "SOMETHING"
        element_to_check = "SOMETHING_ELSE"

        self.assertFalse(_element_matches(element_to_remove, element_to_check))

    def test_element_matches_with_number_arg(self):
        element_to_remove = "SOMETHING"
        element_to_check = "SOMETHING=5"

        self.assertTrue(_element_matches(element_to_remove, element_to_check))

    def test_element_no_match_with_number_arg(self):
        element_to_remove = "SOMETHING"
        element_to_check = "SOMETHING_DIFFERENT=5"

        self.assertFalse(_element_matches(element_to_remove, element_to_check))


class TestRemoveAttributeElement(TestCase):
    def test_remove_element_without_numbers(self):
        current_attribute_state = {"attribute_1": ["ONE", "TWO=2", "THREE"]}
        elements_to_remove = ["ONE", "THREE"]
        expected_result = {"attribute_1": ["TWO=2"]}
        result = _remove_attribute_element(current_attribute_state, "attribute_1", elements_to_remove)

        self.assertEqual(result, expected_result)

    def test_remove_element_with_numbers(self):
        current_attribute_state = {"attribute_1": ["ONE", "TWO=2", "THREE"]}
        elements_to_remove = ["TWO"]
        expected_result = {"attribute_1": ["ONE", "THREE"]}
        result = _remove_attribute_element(current_attribute_state, "attribute_1", elements_to_remove)

        self.assertEqual(result, expected_result)



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from mbed_tools.build.build import build_project, generate_build_system
from mbed_tools.build.exceptions import MbedBuildError


class TestBuildProject(TestCase):
    @mock.patch("mbed_tools.build.build._cmake_wrapper")
    def test_invokes_cmake_with_correct_args(self, cmake_wrapper):
        build_project(build_dir="cmake_build", target="install")

        cmake_wrapper.assert_called_once_with("--build", "cmake_build", "--target", "install")

    @mock.patch("mbed_tools.build.build._cmake_wrapper")
    def test_invokes_cmake_with_correct_args_if_no_target_passed(self, cmake_wrapper):
        build_project(build_dir="cmake_build")

        cmake_wrapper.assert_called_once_with("--build", "cmake_build")

    def test_raises_build_error_if_build_dir_doesnt_exist(self):
        with TemporaryDirectory() as tmp_dir:
            nonexistent_build_dir = pathlib.Path(tmp_dir, "cmake_build")

            with self.assertRaises(MbedBuildError):
                build_project(nonexistent_build_dir)


@mock.patch("mbed_tools.build.build._cmake_wrapper")
class TestConfigureProject(TestCase):
    def test_invokes_cmake_with_correct_args(self, cmake_wrapper):
        source_dir = "source_dir"
        build_dir = "cmake_build"
        profile = "debug"

        generate_build_system(source_dir, build_dir, profile)

        cmake_wrapper.assert_called_once_with(
            "-S", source_dir, "-B", build_dir, "-GNinja", f"-DCMAKE_BUILD_TYPE={profile}"
        )



#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for parsing the attributes for targets in targets.json that override."""
from unittest import TestCase, mock

from mbed_tools.targets._internal.targets_json_parsers.overriding_attribute_parser import (
    get_overriding_attributes_for_target,
    get_labels_for_target,
    _targets_override_hierarchy,
    _determine_overridden_attributes,
    _remove_unwanted_attributes,
    MERGING_ATTRIBUTES,
)
from mbed_tools.targets._internal.targets_json_parsers.accumulating_attribute_parser import ALL_ACCUMULATING_ATTRIBUTES


class TestGetOverridingAttributes(TestCase):
    @mock.patch(
        "mbed_tools.targets._internal.targets_json_parsers." "overriding_attribute_parser._targets_override_hierarchy"
    )
    @mock.patch(
        "mbed_tools.targets._internal.targets_json_parsers."
        "overriding_attribute_parser._determine_overridden_attributes"
    )
    def test_correctly_calls(self, _determine_overridden_attributes, _targets_override_hierarchy):
        target_name = "Target_Name"
        all_targets_data = {target_name: {"attribute_1": ["something"]}}
        result = get_overriding_attributes_for_target(all_targets_data, target_name)

        _targets_override_hierarchy.assert_called_once_with(all_targets_data, target_name)
        _determine_overridden_attributes.assert_called_once_with(_targets_override_hierarchy.return_value)
        self.assertEqual(result, _determine_overridden_attributes.return_value)


class TestParseHierarchy(TestCase):
    def test_overwrite_hierarchy_single_inheritance(self):
        all_targets_data = {
            "D": {"attribute_1": ["some things"]},
            "C": {"inherits": ["D"], "attribute_2": "something else"},
            "B": {},
            "A": {"inherits": ["C"], "attribute_3": ["even more things"]},
        }
        result = _targets_override_hierarchy(all_targets_data, "A")

        self.assertEqual(result, [all_targets_data["A"], all_targets_data["C"], all_targets_data["D"]])

    def test_overwrite_hierarchy_multiple_inheritance(self):
        all_targets_data = {
            "F": {"attribute_1": "some thing"},
            "E": {"attribute_2": "some other thing"},
            "D": {"inherits": ["F"]},
            "C": {"inherits": ["E"]},
            "B": {"inherits": ["C", "D"]},
            "A": {"inherits": ["B"]},
        }
        result = _targets_override_hierarchy(all_targets_data, "A")

        self.assertEqual(
            result,
            [
                all_targets_data["A"],
                all_targets_data["B"],
                all_targets_data["C"],
                all_targets_data["E"],
                all_targets_data["D"],
                all_targets_data["F"],
            ],
        )


class TestOverridingAttributes(TestCase):
    def test_determine_overwritten_attributes(self):
        override_order = [
            {"attribute_1": "1"},
            {"attribute_1": "I should be overridden", "attribute_2": "2"},
            {"attribute_3": "3"},
        ]
        expected_attributes = {"attribute_1": "1", "attribute_2": "2", "attribute_3": "3"}

        result = _determine_overridden_attributes(override_order)
        self.assertEqual(result, expected_attributes)

    def test_remove_accumulating_attributes(self):
        override_order = [
            {ALL_ACCUMULATING_ATTRIBUTES[0]: "1"},
            {"attribute": "Normal override attribute"},
            {ALL_ACCUMULATING_ATTRIBUTES[1]: "3"},
        ]
        expected_attributes = {"attribute": "Normal override attribute"}

        result = _determine_overridden_attributes(override_order)
        self.assertEqual(result, expected_attributes)

    def test_merging_attributes(self):
        override_order = [
            {MERGING_ATTRIBUTES[0]: {"FOO": "I should also remain"}},
            {MERGING_ATTRIBUTES[0]: {"FOO": "I should not remain"}},
            {MERGING_ATTRIBUTES[0]: {"BAR": "I should remain"}},
        ]
        expected_attributes = {MERGING_ATTRIBUTES[0]: {"BAR": "I should remain", "FOO": "I should also remain"}}

        result = _determine_overridden_attributes(override_order)
        self.assertEqual(result, expected_attributes)


class TestGetLabelsForTarget(TestCase):
    def test_linear_inheritance(self):
        all_targets_data = {
            "C": {"attribute_1": ["some things"]},
            "B": {"inherits": ["C"], "attribute_2": "something else"},
            "A": {"inherits": ["B"], "attribute_3": ["even more things"]},
        }
        target_name = "A"
        expected_result = {"A", "B", "C"}
        result = get_labels_for_target(all_targets_data, target_name)

        self.assertSetEqual(result, expected_result)

    def test_multiple_inheritance(self):
        all_targets_data = {
            "E": {"attribute_1": "some thing"},
            "D": {"attribute_2": "some other thing"},
            "C": {"inherits": ["E"]},
            "B": {"inherits": ["C", "D"]},
            "A": {"inherits": ["B"]},
        }
        target_name = "A"
        expected_result = {"A", "B", "C", "D", "E"}
        result = get_labels_for_target(all_targets_data, target_name)

        self.assertSetEqual(result, expected_result)

    def test_no_inheritance(self):
        all_targets_data = {
            "A": {"attribute_3": ["some things"]},
        }
        target_name = "A"
        expected_result = {"A"}
        result = get_labels_for_target(all_targets_data, target_name)

        self.assertSetEqual(result, expected_result)


class TestRemoveUnwantedAttributes(TestCase):
    def test_removes_accumulating_public_and_inherits(self):
        target_attributes = {
            ALL_ACCUMULATING_ATTRIBUTES[0]: "1",
            "public": False,
            "inherits": "SomeOtherBoard",
            "attribute": "I should remain",
        }
        expected_result = {"attribute": "I should remain"}

        result = _remove_unwanted_attributes(target_attributes)
        self.assertEqual(result, expected_result)
